#######################################################################
# This file is part of Lyntin.
# copyright (c) Free Software Foundation 2001-2007
#
# Lyntin is distributed under the GNU General Public License license.  See the
# file LICENSE for distribution details.
# $Id: textui.py,v 1.10 2007/07/24 00:39:03 willhelm Exp $
#######################################################################
"""
Holds the text ui class.
"""
import re, sys, os, select
from lyntin import ansi, engine, event, utils, exported, config
from lyntin.ui import base, message


HELP_TEXT = """
The textui is the most basic ui you can get.  It works great over
telnet/ssh, but terrible in the Win32 command window.  At the same 
time, because it's so basic, it tends to be a good testing ui.

To enable GNU readline support (if it's available) use:

   --readline on

at the command line.  It'll look for a "readlinerc" file in your
datadir and set readline options from that.

Also, depending on what terminal you're running Lyntin from, it might 
help to set the TERMTYPE at the commandline.  Lyntin returns "lyntin" 
when asked, but doing something like:

   --term xterm

will kick Lyntin into returning "xterm" as the TERMTYPE.
"""

try:
  import termios
except ImportError:
  pass

myui = None

DEFAULT_COLOR = list(ansi.DEFAULT_COLOR)
DEFAULT_ANSI = chr(27) + "[0m"

def get_ui_instance():
  global myui
  if myui == None:
    myui = Textui()
  return myui

class Textui(base.BaseUI):
  """
  This is the text ui.  It's super basic and should run almost
  anywhere.  It lacks several useful functions that the Tkui
  and the Curses ui (eventually) will have.
  """
  def __init__(self):
    """ Initialize the textui."""
    base.BaseUI.__init__(self)
    exported.hook_register("shutdown_hook", self.shutdown)
    exported.hook_register("to_user_hook", self.write)
    exported.hook_register("config_change_hook", self.configChangeHandler)
    exported.hook_register("bell_hook", self.bell)
    self._currcolors = {}
    self._unfinishedcolor = {}

    self._tio = 0
    self._rline = 0
    self._echo = 1

    # grab the stdin stream at once,
    # enabling it's further redirection by another module
    self._stdin = sys.stdin

  def runui(self):
    global HELP_TEXT
    exported.add_help("textui", HELP_TEXT)
    exported.write_message("For textui help, type \"#help textui\".")

    # termios is the module that allows us to change echo for a terminal
    # but only if the module is present
    try:
        import termios
    except ImportError:
        self._tio = 0
    else:
        self._tio = 1
        echonew = termios.tcgetattr(self._stdin.fileno())
        self._onecho_attr = echonew[3]
        self._offecho_attr = echonew[3] & ~termios.ECHO

    if "readline" in config.options:
      try:
        import readline
      except ImportError:
        self._rline = 0
        exported.write_error("Readline not available for your system.")
      else:
        self._rline = 1

        # we do some stuff to grab the readlinerc file if they have one
        # so the user can set some readline oriented things which makes 
        # things a little nicer for the user.
        d = exported.get_config("datadir")

        try:
          readline.read_init_file(d + "readlinerc")
        except FileNotFoundError:
          exported.write_error(f"Note: No readlinerc file available in {d}.")
        except Exception as e:
          exported.write_error(f"Error loading readlinerc file in {d}: {e}")
        
        exported.write_message("Readline enabled.")

    if self._tio == 0 or self._rline == 1:
      exported.write_error("Warming: echo off is unavailable.  " +
                           "Your password will be visible.")

    # go into the main loop
    self.run()
      
  def shutdown(self, args):
    """ Shuts down the textui and makes sure that we're echoing!"""
    if self._echo == 0:
      self.turnonecho()

  def turnonecho(self):
    """ Turns on echo if termios module is present."""
    if self._tio == 0 or self._rline == 1:
      return

    # Ensure stdin is valid and supports fileno()
    if not self._stdin or not hasattr(self._stdin, "fileno"):
      print("Error: stdin is invalid or does not support fileno()")
      return

    # Ensure we're in a real terminal
    if not sys.stdin.isatty():
      print("Error: Not running in a real terminal.")
      return

    try:
      fd = self._stdin.fileno()
      new = termios.tcgetattr(fd)
      new[3] = self._onecho_attr
      termios.tcsetattr(fd, termios.TCSADRAIN, new)
      self._echo = 1
    except termios.error as e:
      exported.write_error(f"textui: unable to turn on echo: {e}")

  def turnoffecho(self):
    """ Turns off echo if termios module is present."""
    if self._tio == 0 or self._rline == 1:
      return

    fd = self._stdin.fileno()
    new = termios.tcgetattr(fd)
    new[3] = self._offecho_attr
    try:
      termios.tcsetattr(fd, termios.TCSADRAIN, new)
      self._echo = 0
    except Exception as e:
      exported.write_error("textui: unable to turn off echo: %s" % e)


  def bell(self, args):
    """ Handles incoming bell characters."""
    sys.stdout.write('\07')

  def configChangeHandler(self, args):
    """ Handles config changes (including mudecho)."""
    name = args["name"]
    newvalue = args["newvalue"]

    if name == "mudecho":
      if newvalue == 0:
        # echo off
        self.turnoffecho()
      else:
        # echo on
        self.turnonecho()

  def _posix_readline_input(self):
    """
    If the os is posix and the readline module is present, then we 
    use input to grab user input.
    """
    try:
      return eval(input())
    except EOFError:
      pass

  def _posix_input(self):
    """
    If the os is posix and there is no readline module, then we
    use sys.stdin.readline().
    """
    readers,w,e = select.select([self._stdin], [], [])
    if readers:
      for mem in readers:
        try:
          return mem.readline()
        except IOError:
          pass

  def _non_posix_input(self):
    return self._stdin.readline()

  def run(self):
    """ This is the poll loop for user input."""
    try:
      while not self.shutdownflag:
        if os.name == 'posix':
          if self._rline == 1:
            data = self._posix_readline_input()
          else:
            data = self._posix_input()
        else:
          data = self._non_posix_input()

        if data != None:
          self.handleinput(data)

          # FIXME - this is just plain icky.  the issue is that
          # we need to know we're ending _before_ we block for
          # the next input.  otherwise Lyntin will shut down except
          # for this thread which will hang around blocking until
          # the user hits the enter key.
          # 
          # any good ideas for dealing with this are more than welcome.
          if data.find("#end") == 0:
            break


    except select.error as e:
      (errno,name) = e
      if errno == 4:
        exported.write_message("system exit: select.error.")
        event.ShutdownEvent().enqueue()
        return

    except SystemExit:
      exported.write_message("system exit: you'll be back...")
      event.ShutdownEvent().enqueue()

    except:
      exported.write_traceback()
      event.ShutdownEvent().enqueue()


  def write(self, args):
    """
    Handles writing information from the mud and/or Lyntin
    to the user.
    """
    msg = args["message"]

    if isinstance(msg, str):
      msg = message.Message(msg, message.LTDATA)

    line = msg.data
    ses = msg.session

    if line == '' or self.showTextForSession(ses) == 0:
      return

    # we prepend the session name to the text if this is not the 
    # current session sending text.
    pretext = ""
    if ses != None and ses != exported.get_current_session():
      pretext = "[" + ses.getName() + "] "

    if msg.type == message.ERROR or msg.type == message.LTDATA:
      if msg.type == message.ERROR:
        pretext = "error: " + pretext
      else:
        pretext = "lyntin: " + pretext

      line = pretext + utils.chomp(line).replace("\n", "\n" + pretext)
      if exported.get_config("ansicolor") == 1:
        line = DEFAULT_ANSI + line
      sys.stdout.write(line + "\n")
      return

    elif msg.type == message.USERDATA:
      # we don't print user data in the textui
      return

    if exported.get_config("ansicolor") == 0:
      if pretext:
        if line.endswith("\n"):
          line = (pretext + line[:-1].replace("\n", "\n" + pretext) + "\n")
        else:
          line = pretext + line.replace("\n", "\n" + pretext)
      sys.stdout.write(line)
      sys.stdout.flush()
      return

    # each session has a saved current color for mud data.  we grab
    # that current color--or user our default if we don't have one
    # for the session yet.
    if ses in self._currcolors:
      color = self._currcolors[ses]
    else:
      # need a copy of the list and not a reference to the list itself.
      color = list(DEFAULT_COLOR)


    # some sessions have an unfinished color as well--in case we
    # got a part of an ansi color code in a mud message, and the other
    # part is in another message.
    if ses in self._unfinishedcolor:
      leftover = self._unfinishedcolor[ses]
    else:
      leftover = ""

    lines = line.splitlines(1)
    if lines:
      for i in range(0, len(lines)):
        mem = lines[i]
        acolor = ansi.convert_tuple_to_ansi(color)

        color, leftover = ansi.figure_color(mem, color, leftover)

        if pretext:
          lines[i] = DEFAULT_ANSI + pretext + acolor + mem
        else:
          lines[i] = DEFAULT_ANSI + acolor + mem

      sys.stdout.write("".join(lines) + DEFAULT_ANSI)
      sys.stdout.flush()

    self._currcolors[ses] = color
    self._unfinishedcolor[ses] = leftover


  def prompt(self):
    """ Prints a prompt to the user."""
    sys.stdout.write("> ")
    sys.stdout.flush()

  def flush(self):
    """ Flushes the stdout.  Not sure we really need this
    but it's here."""
    sys.stdout.flush()

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: lyntincmds.py,v 1.11 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This module holds commands that are new and unique to Lyntin.
"""
import types, re
from lyntin import net, utils, engine, constants, exported, config
from lyntin.modules import modutils

commands_dict = {}

def _fixmap(w, themap):
  keys = sorted(themap.keys())
  output = []

  for mem in keys:
    if isinstance(themap[mem], list):
      if len(themap[mem]) > 0:
        output.append(f"   {mem.ljust(w)} {themap[mem][0]}")  # Use f-strings
        for mem2 in themap[mem][1:]:
          output.append((18 * " ") + mem2)
      else:
        output.append("   %s (None)" % (mem.ljust(w)))
    else:
      output.append("   %s %s" % (mem.ljust(w), themap[mem]))
  return "\n".join(output)


def bv(bool):
  if bool:
    return "on"
  return "off"

def chr_cmd(ses, args, input):
  """
  Allows you to assign arbitrary characters to variables.  For example,
  if you wanted to assign ASCII char 7 to variable ctrlG you could
  do:

    #chr {ctrlG} {7}

  Since this creates a variable, you should remove the variable with
  the unvariable command.

  Note: This won't work if you don't have the variable module loaded.

  category: commands
  """
  var = args["var"]
  ascii = args["ascii"]
  quiet = args["quiet"]

  vm = exported.get_manager("variable")

  if not vm:
    exported.write_error("chr: no variable manager found.")
    return

  if ascii < 0 or ascii > 127:
    exported.write_error("chr: ascii argument out of range of 0 to 127.")
    return

  vm.addVariable(ses, var, chr(ascii))
  if not quiet:
    exported.write_message("chr: variable %s added." % var)

commands_dict["chr"] = (chr_cmd, "var ascii:int quiet:boolean=false")


def config_cmd(ses, args, input):
  """
  Allows you to view and change configuration options that affect
  how Lyntin functions.  Configuration options can be session
  oriented or global to all of Lyntin.

  examples: 
    #config
        displays global configuration and session configuration for the 
        current session

    #a #config
        displays global configuration and session configuration for the 
        session named 'a'

    #config ansicolor
        displays information about the mudecho configuration option

    #config ansicolor on
        sets the ansicolor configuration option to on

  category: commands
  """
  name = args["name"]
  value = args["value"]
  quiet = args["quiet"]

  c = exported.myengine.getConfigManager()

  # if they didn't specify a name, then we print out all the
  # configuration stuff for general and this session
  if not name:
    general = c.getConfigItems(None)
    globmap = {}
    for mem in general:
      globmap[mem._name] = mem.toString()

    seslisting = c.getConfigItems(ses)
    sesmap = {}
    for mem in seslisting:
      sesmap[mem._name] = mem.toString()

    output = "Commandline:\n"
    output += _fixmap(16, config.options) + "\n"

    output += "Global:\n"
    output += _fixmap(16, globmap) + "\n"

    output += "Session:\n"
    output += _fixmap(16, sesmap) + "\n"

    exported.write_message(output, ses)
    return

  # try to find a session item first
  ci = c.getConfigItem(name, ses)
  if not ci:
    ci = c.getConfigItem(name)

  if not ci:
    exported.write_error("config: config manager does not recognize %s as a config item." % name)
    return


  if not value:
    # we print out everything we know about this config item.
    output = "config: %s\n\ncurrent value: %s\n\n%s\n" % \
             (name, ci.toString(), utils.wrap_text(ci._description, wraplength=60))
    exported.write_message(output)
    return


  try:
    try:
      c.change(name, value, ses)
    except:
      c.change(name, value)
    exported.write_message("config: %s set to %s." % (name, value), ses)
  except Exception as e:
    exported.write_error(e)

commands_dict["config"] = (config_cmd, "name= value= quiet:boolean=false")
  
def grep_cmd(ses, args, input):
  """
  Similar to the unix grep command, this allows you to extract 
  information from the session's data buffer using regular expressions.

  It prints matching lines in their entirety.

  examples:
    #grep {says:} 1000

    Greps the last 1000 lines of the databuffer for lines that have
    "says:" in them.

  category: commands
  """
  if (ses.getName() == "common"):
    exported.write_error("grep cannot be applied to common session.", ses)
    return

  pattern = args["pattern"]
  size = args["size"]
  context = args["context"]
  buffer = ses.getDataBuffer()

  ret = []
  cpattern = re.compile(pattern)
  for i in range(max(len(buffer)-size,0), len(buffer)):
    mem = buffer[i]
    if cpattern.search(mem):
      if context > 0:
        mem = []
        if i > 0:
          bound = i - context
          if bound < 0: bound = 0
          for j in range(bound, i):
            mem.append("  " + buffer[j])

        mem.append("+ " + buffer[i])

        if i < len(buffer):
          bound = i+context+1
          if bound > len(buffer)-1: bound = len(buffer)-1
          for j in range(i+1, bound):
            mem.append("  " + buffer[j])
        mem = "".join(mem)

      ret.append(mem)

  if context == 0:
    splitter = ""
  else:
    splitter = "---\n"
  exported.write_message("grep %s results:\n%s" % (pattern, splitter.join(ret)), ses)

commands_dict["grep"] = (grep_cmd, "pattern size:int=300 context:int=0")


def diagnostics_cmd(ses, args, input):
  """
  This is very useful for finding out all the information about Lyntin
  while it's running.  This will print out operating system information,
  Python version, what threads are running (assuming they're registered
  with the ThreadManager), hooks, functions connected to hooks, and
  #info for every session.  It's very helpful in debugging problems that
  are non-obvious or are platform specific.  It's also invaluable in
  bug-reporting.

  It can take a filename argument and will copy the #diagnostics output
  to that file.  This allows you easier method of submitting diagnostics
  output along with bug reports.

  Note: Windows users should either use two \\'s or use / to separate
  directory names.

  category: commands
  """
  import os, sys
  message = []
  message.append("Diagnostics:")
  message.append(exported.myengine.getDiagnostics())

  message.append("Hook statii:")
  data = exported.myengine.checkHooks()
  data.sort()
  for mem in data:
    message.append(mem)

  message.append("Thread statii:")
  data = exported.myengine.checkthreads()
  data.sort()
  for mem in data:
    message.append(mem)
      
  message.append("OS/Python information:")
  try: 
    message.append("   sys.version: %s" % sys.version)
  except:
    message.append("   sys.version not available.")

  try: 
    message.append("   os.name: %s" % os.name)
  except:
    message.append("   os.name not available.")
 
  message.append("   lyntin: %s" % (constants.VERSION[:constants.VERSION.find("\n")]))

  message.append("Lyntin Options:")
  for mem in list(config.options.keys()):
    message.append("   %s: %s" % (mem, repr(config.options[mem])))

  exported.write_message("\n".join(message))
  exported.write_message("This information can be dumped to a "
        "file by doing:\n   #diagnostics dumpfile.txt")

  logfile = args["logfile"]
  if logfile:
    import time
    try:
      f = open(logfile, "w")
      f.write("This file was created on: %s" % time.asctime())
      f.write(os.linesep + os.linesep)
      f.write(os.linesep.join(message))
      f.close()
      exported.write_message("diagnostics: written out to file %s." % logfile)
    except Exception as e:
      exported.write_error("diagnostics: Error writing to file %s. %s" 
                            % (logfile, e))

commands_dict["diagnostics"] = (diagnostics_cmd, "logfile=")


def raw_cmd(ses, args, input):
  """
  Sends input straight to the mud.

  category: commands
  """
  if (ses.getName() == "common"):
    exported.write_error("raw: cannot send raw data to the common session.", ses)
    return

  ses.writeSocket(args["input"] + "\n")
  
commands_dict["raw"] = (raw_cmd, "input=", "limitparsing=0")


def load():
  """ Initializes the module by binding all the commands."""
  modutils.load_commands(commands_dict)


def unload():
  """ Unloads the module by calling any unload/unbind functions."""
  modutils.unload_commands(list(commands_dict.keys()))

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#######################################################################
# This file is part of Lyntin.
# copyright (c) Free Software Foundation 2001-2007
#
# Lyntin is distributed under the GNU General Public License license.  See the
# file LICENSE for distribution details.
# $Id: base.py,v 1.10 2007/07/24 00:39:03 willhelm Exp $
#######################################################################
"""
Holds the base ui class for Lyntin as well as the get_ui function
"""
import string, re, sys, os
from lyntin import exported, utils
from lyntin.ui import message


def get_ui(uiname):
  """
  Attempts to retrieve the ui by that name.

  @param uiname: the name of the ui passed in by the command line
  @type  uiname: string

  @return: a BaseUI subclass instance corresponding to the name
      of the ui the user wants to instantiate.  or None if the ui
      could not be found or instantiated.
  @rtype: BaseUI subclass
  """
  path = "lyntin.ui." + uiname
  try:
    _module = __import__(path)
    _module = sys.modules[path]
  except ImportError:
    raise ValueError("ui '%s' does not exist." % uiname)
  return _module.get_ui_instance()

class BaseUI:
  """
  Base ui class.

  This is the Base UI class which defines the interface between
  the ui's and Lyntin.
  """
  def __init__(self):
    """
    If you have initializations to do, override this class,
    but call this function like this::

      BaseUI.__init__(self)

    then go on to do the initializing you need to do.
    """
    self.shutdownflag = 0
    self.reset_completion()
    import lyntin.exported
    lyntin.exported.hook_register("shutdown_hook", self.shutdown)

  def runui(self):
    """
    The engine executes this expecting your mainloop to start and
    you to take control of the main thread of execution.  Put
    whatever other initialization you need to do here and then 
    go into your main loop.
    """
    pass

  def wantMainThread(self):
    """
    This should return 0 (if you don't want the main thread of
    execution for the ui) or a 1 (if you do want the main thread
    of execution for the ui).  This defaults to 0.

    @returns: 0 or 1
    @rtype: boolean
    """
    return 0

  def write(self, args):
    """
    Writes output to the user.  Output can come from the mud, 
    Lyntin, or even user input being printed to the screen.  If 
    the message argument is a String object rather than a Message
    object, the ui should assume it's Lyntin output.

    This method should be registered with the to_user_hook.

    @param args: either a string or a Message instance--this is
        the thing to be outputted to the user
    @type args: tuple holding either a string or a Message instance
    """
    pass

  def prompt(self):
    """
    Prints a prompt to the user.  This is mostly for niceties so the 
    user knows that Lyntin is awaiting input.  It should just print
    a prompt.  Prompts only get printed by the common session.
    """
    pass

  def shutdown(self, args):
    """
    Tells the user interface thread to shutdown.  This is 
    registered with the shutdown_hook.  Implement this method
    if you have shutdown stuff to do.

    @param args: it's an empty tuple--we ignore this
    @type  args: tuple of 0 length
    """
    self.shutdownflag = 1

  def flush(self):
    """
    Flushes output to the user.  Currently we don't do any flushing
    and it shouldn't be needed.  I'm not wholly sure why it's here.
    """
    pass

  def showTextForSession(self, ses):
    """
    Returns whether or not we should show text for this session--it's
    a convenience method.

    We return a 1 if the session is None, it doesn't have a _snoop
    attribute, it's the current session, or get_config("snoop", ses) == 1.

    @param ses: the session we're looking at--if it's None we return a 1
    @type  ses: Session

    @returns: 1 if we should show text, 0 if not
    @rtype: boolean
    """
    if ses == None or getattr(ses, "_snoop", None) == None \
        or exported.get_current_session() == ses \
        or exported.get_config("snoop", ses, 1) == 1:
      return 1
    return 0

  def handleinput(self, input):
    """
    Nicely handles enqueuing of input events.  Also deals with things 
    like CR and LF.  Call this method with input that's just been
    polled from the user.

    @param input: the raw input from the user
    @type  input: string
    """
    from lyntin import event
    input = utils.chomp(input)
    event.InputEvent(input).enqueue()

  def get_completion(self, text='', position=None):
    """
    This method should be called when some hotkey (usually "Tab", but it
    is up to the UI module) is being pressed.

    @param text: the text to be completed
    @type  text: string

    @param position: cursor position in the text
    @type  position: int

    @returns: next possible completion pair (newtext, newposition)
    @rtype:   (string, int)
    """
    complist = self.get_completion_list(text, position)
    completion = complist[0]
    complist[:1] = []
    complist.append(completion)
    return completion

  def get_completion_list(self, text='', position=None):
    """
    Returns the completion list. Rebuilds the list if it is empty.

    @param text: the text to be completed
    @type  text: string

    @param position: cursor position in the text
    @type  position: int
    
    @returns: all possible completion pairs list [(newtext, newposition)]
    @rtype:   [(string, int)]
    """
    if not self.completion_list_:
      #
      # Rebuild the completion list if it is empty after reset.
      #
      if position==None or position<0 or position>len(text):
        position = len(text)
      #
      # Gather all the completions from registered completers:
      #
      exported.hook_spam("completer_hook",
                         { 'text': text, 'position': position },
                         self._completion_mapping_helper
                        )
      # put an anchor at the end of list
      self.completion_list_.append((text, position))
    return self.completion_list_

  def reset_completion(self):
    """
    Resets the completion in progress. The next call to "get_completion"
    will reinitialize the completion list.
    """
    self.completion_list_ = []
    
  def _completion_mapping_helper(self, x, y):
    self.completion_list_ += y
    return x

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: session.py,v 1.14 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
Holds the functionality involved in X{session}s.  Sessions are copied 
from the common session.  Each session encapsulates a socket connection
to a mud--though it should be noted that sessions could also connect
to any other TCP/IP service.

X{disconnect_hook}::

   When the connection from Lyntin to the mud ends, we spam this hook.

   Arg mapping: { "session": Session, "host": string, "port": int }

   session - the session that we just lost the connection to

   host - the mud we were connected to

   port - the port we were connected to


X{to_mud_hook}::

   All data that gets sent from Lyntin to the mud gets spammed
   to this hook.  This is different from the from_user_hook
   because this data has already passed through Lyntin's 
   transform mechanisms.

   Arg mapping: { "session": Session, "data": string, "tag": varies }

   session - the session of the mud we're sending the data to

   data - the raw data we're sending

   tag - this allows you to "tag" outgoing data so that you can 
         correlate the outgoing data with incoming data and build
         lock-step mechanisms that track user input and mud output.


X{user_filter_hook}::

   After data has passed on the from_user_hook it gets passed through
   the user_filter_hook which allows the data from the user to
   be transformed.  This includes alias expansion, variable expansion,
   speedwalk expansion and whatever other modules listen to this
   hook.

   Functions that register with this hook should return the dataadj
   if they did nothing or the adjusted dataadj if they transformed it.
   Look at examples in the alias and speedwalk modules.

   Arg mapping: { "session": Session, "internal": boolean, "verbatim": boolean,
                  "data": string, "dataadj": string }

   session - the session this data is associated with

   internal - whether or not this is internally generated.  this affects
              whether we spam the from_user_hook and whether this 
              gets logged in the HistoryManager.

   verbatim - whether or not we should be transforming this user data item.

   data - the original raw user data

   dataadj - the latest adjusted data (which originally was the original
             raw user data)


X{mud_filter_hook}::

   Data that comes from the mud gets passed through the from_mud_hook first.
   Then it passes through the mud_filter_hook where it gets transformed.
   Actions, substitutes, gags, and things of that nature should register
   with this hook.

   Functions that register with this hook should return the dataadj if
   they don't want to adjust the mud data or the adjusted dataadj if they
   do adjust the mud data.  See the action, gag and highlight modules for
   examples.
   
   Arg mapping: { "session": Session, "data": string, "dataadj": string }

   session - the Session associated with the mud this data came from

   data - the original raw data from the mud

   dataadj - the latest adjusted data from the mud
"""
import re, copy, string, os
from lyntin import exported, utils, ansi, event

ESC = chr(27)

class Session:
  """
  A session is a nice container of all the stuff that encompasses a 
  user session: aliases, actions, commands...

  All input and output goes through the Session object.
  Almost everything happens through the Session.
  """
  global_vars = {}

  def __init__(self, engine_instance):
    """
    Initialize.
    """
    self._engine = engine_instance
    self._socket = None
    self._name = ""
    self._host = "none"
    self._port = 0
    self._colorbuffer = ''

    self._databuffer = []
    self._databuffersize = 10000

    # register with the shutdown hook 
    self._engine.hookRegister("shutdown_hook", self.shutdown)

    # session variables
    self._vars = {}

  def __repr__(self):
    return "session.Session %s" % self._name

  def setupCommonSession(self):
    from lyntin import config
    c = self._engine.getConfigManager()

    if type(config.options["snoopdefault"]) is list:
      config.options["snoopdefault"] = config.options["snoopdefault"][0]

    tc = config.BoolConfig("snoop", utils.convert_boolean(config.options['snoopdefault']), 0,
          "Whether or not you see the data from this session when it's not "
          "active.")
    c.add("snoop", tc, self)
    
    tc = config.BoolConfig("verbatim", 0, 0,
          "Whether we're in verbatim mode where we pass the user text "
          "straight to the mud without massaging it.")
    c.add("verbatim", tc, self)

  def getName(self):
    """
    Returns the name of the session.

    @returns: the name of the session
    @rtype: string
    """
    return self._name

  def setName(self, name):
    """
    Sets the name of the session.

    @param name: the new name
    @type  name: string
    """
    self._name = name
    if self._socket:
      self._socket.setSessionName(name)

  def shutdown(self, args):
    """
    Shuts down the session, shuts down the underlying SocketCommunicator.

    @param args: the args tuple for the shutdown_hook.
    @type  args: tuple
    """
    if len(args) > 0:
      quiet = 1
    else:
      quiet = 0

    # unregister with the shutdown hook
    # exported.hook_unregister("shutdown_hook", self.shutdown)
    if self.getName() != "common":
      if quiet == 0:
        event.OutputEvent("Session %s disconnected.\n\"#zap %s\" to kill the session.\n" % (self._name, self._name)).enqueue()
      exported.hook_spam("disconnect_hook", {"session": self, "host": self._host,  "port": self._port})

      if self._socket:
        self._socket.shutdown()

      self._host = None
      self._port = 0
      self._socket = None

  def getStatus(self):
    """
    Returns status of the session.  Most specifically the session name,
    the socket we're connected to.

    @returns: the status string
    @rtype: string
    """
    data = []
    
    data.append("Session name: %s" % self._name)
    data.append("   socket: %s" % repr(self._socket))

    return data

  def clear(self):
    """
    Clears the session (except for connections).  Goes through
    the list of managers registered with the engine and calls
    the clear method with itself.
    """
    for mem in list(self._engine._managers.values()):
      mem.clear(self)

    self._databuffer = []


  ### ------------------------------------------------
  ### Socket stuff
  ### ------------------------------------------------

  def setSocketCommunicator(self, sc):
    """
    Sets the socket communicator.

    @param sc: the new SocketCommunicator instance
    @type  sc: SocketCommunicator
    """
    self._socket = sc

  def getSocketCommunicator(self):
    """
    Returns the socket communicator.

    @returns: SocketCommunicator instance
    @rtype: SocketCommunicator
    """
    return self._socket

  def isConnected(self):
    """
    Tells you whether (1) or not (0) a session has a connection.

    @returns: 1 if connected, 0 if not
    @rtype: boolean
    """
    return self._socket is not None

  def writeSocket(self, message, tag=None):
    """
    Writes data to the socket.

    @param message: the data to be written to the mud
    @type  message: string

    @param tag: Used to tag data being sent to the mud for
        identification when it comes out of the to_mud_hook.
        Simply passed through as-is by lyntin internals.
    @type  tag: varies
    """
    for line in message.strip().split("\n"):
      exported.hook_spam("to_mud_hook", {"session": self, "data": line, "tag": tag})

    if self._socket:
      retval = self._socket.write(str(message))
      if retval:
        exported.write_error("socket write: %s" % retval)

    else:
      # if we don't have a socket then we can't do any non-lyntin-command
      # stuff.
      exported.write_error("No connection.  Create a session.\n(See also: #help, #help session)")
      return


  ### ------------------------------------------------
  ### Data buffer stuff
  ### ------------------------------------------------
  def getDataBuffer(self):
    """
    Returns the DataBuffer instance for this session.

    @returns: list of strings
    @rtype: list of strings
    """
    return self._databuffer

  def addToDataBuffer(self, text):
    """
    Adds data to the buffer by thinking about everything
    in terms of lines.
    
    @param text: the text to add to the buffer
    @type  text: string
    """
    text = ansi.filter_ansi(utils.filter_cm(text))
    lines = text.splitlines(1)

    for mem in lines:
      if len(self._databuffer) == 0 or self._databuffer[-1].endswith("\n"):
        self._databuffer.append(mem)
      else:
        self._databuffer[-1] += mem

    if len(self._databuffer) > self._databuffersize:
      self._databuffer[:-self._databuffersize] = []

  def clearDataBuffer(self):
    """ 
    Clears the databuffer.
    """
    self._databuffer = []
  
  def resizeDataBuffer(self, newsize=10000):
    """ 
    Changes the buffer max.

    @param newsize: the new buffer max size
    @type  newsize: int
    """
    self._databuffersize = newsize


  ### ------------------------------------------------
  ### Variable managing stuff
  ### ------------------------------------------------
  def _varChangeHook(self, var, old, new):
    """
    This calls the variable_change_hook.  It allows other modules to
    know when variable values are changed so that they can handle
    those changes accordingly.
    """
    exported.hook_spam("variable_change_hook", 
          {"session": self, "variable": var, "oldvalue": old, "newvalue": new})

  def setVariable(self, var, expansion):
    """
    Sets a variable value.  This also calls the variable_change_hook.

    @param var: the variable name to set
    @type  var: string

    @param expansion: the new value of the variable.  this can be a string
        or a class with a __str__ method.
    @type  expansion: string
    """
    if var.startswith("_"):
      d = Session.global_vars
    else:
      d = self._vars

    oldvalue = d.get(var, None)
    d[var] = expansion

    self._varChangeHook(var, oldvalue, expansion)

  def removeVariable(self, var):
    """
    Removes a variable from the local _vars dict.  This also spams the
    variable_change_hook.

    @param var: the name of the variable to remove
    @type  var: string
    """
    if var.startswith("_"):
      d = Session.global_vars
    else:
      d = self._vars

    if var in d:
      oldvalue = d[var]
      del d[var]
      self._varChangeHook(var, oldvalue, None)

  def getVariable(self, var, default=None):
    if var.startswith("_"):
      d = Session.global_vars
    else:
      d = self._vars

    return d.get(var, default)

  ### ------------------------------------------------
  ### User input functions
  ### ------------------------------------------------
  def prompt(self):
    """
    Deals with printing a prompt if this is the common session.
    """
    if self.getName() == "common":
      self._engine.writePrompt()

  def handleUserData(self, input, internal=0 ):
    """
    Handles input in the context of this session specifically.

    @param input: the user data
    @type  input: string

    @param internal: whether the command came from interally.
        we won't spam hooks and may at some point prevent
        output for internal stuff too.  1 if internal, 0 if not.
    @type  internal: boolean
    """
    # this is the point of much recursion.  everything is registered
    # as a filter and recurses accordingly.
    spamargs = {"session": self, "internal": internal, 
                "verbatim": exported.get_config("verbatim", self), 
                "data": input, "dataadj": input}
    spamargs = exported.filter_mapper_hook_spam("user_filter_hook", spamargs)

    if spamargs is None:
      return
    else:
      input = spamargs["dataadj"]

    # after this point we don't do any more recursion.  so it's
    # safe to unescape things and such.
    input = input.replace("\\;", ";")
    input = input.replace("\\$", "$")
    input = input.replace("\\%", "%")

    # just regular data to the mud
    self.writeSocket(input + "\n")


  ### ------------------------------------------------
  ### Mud input functions
  ### ------------------------------------------------

  def handleMudData(self, input):
    """
    Handles input coming from the mud.

    @param input: the data coming from the mud
    @type  input: string
    """
    # this sort of handles ansi color codes that get broken 
    # mid-transmission when mud data is chunked and sent across
    # the network.
    if self._colorbuffer:
      input = self._colorbuffer + input
      self._colorbuffer = ''

    index = input.rfind(ESC)
    if index != -1 and input.find("m", index) == -1:
      self._colorbuffer = input[index:]
      input = input[:index]

    # we add the new input to the databuffer
    self.addToDataBuffer(input)

    # we split the input into a series of lines and operate on
    # those
    inputlines = input.splitlines(1)

    for i in range(0, len(inputlines)):
      mem = inputlines[i]
      # call the pre-filter hook
      spamargs = {"session": self, "data": mem, "dataadj": mem}

      spamargs = exported.filter_mapper_hook_spam("mud_filter_hook", spamargs)
      if spamargs is not None:
        mem = spamargs["dataadj"]
      else:
        mem = ""

      inputlines[i] = mem

    exported.write_mud_data("".join(inputlines), self)


# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: variable.py,v 1.14 2007/07/24 00:39:03 willhelm Exp $
#########################################################################

"""
This module defines the VariableManager which handles variables.
It also defines global variables like $TIMESTAMP.

X{variable_change_hook}::

   The variable_change_hook allows other modules to be notified when
   variable values change and handle those changes accordingly.  For
   instance, the action module would recompile its regular expression
   triggers based on the new values of the variables.

   Arg mapping: { "session": Session, "variable": string, "oldvalue": string,
                  "newvalue": string }

   session - the session this variable belongs to

   variable - the name of the variable in question

   oldvalue - the old value of the variable

   newvalue - the new value of the variable
"""
import time
from lyntin import manager, utils, config, engine, exported, session
from lyntin.modules import modutils

class DatadirBuiltin:
  """
  Allows us to do dynamic DATADIRs as a global variable.
  """
  def __init__(self): pass
  def __str__(self): return config.options["datadir"]

class TimeStampBuiltin:
  """
  Allows us to do dynamic TIMESTAMPs as a global variable.
  """
  def __init__(self): pass
  def __str__(self): return time.asctime()

class LogTimeStampBuiltin:
  """
  Allows us to do dynamic TIMESTAMPs as a global variable in the
  form yyyymmddhhmmss.  Good for logfiles.
  """
  def __init__(self): pass
  def __str__(self): return time.strftime('%Y%m%d%H%M%S')

class VariableManager(manager.Manager):
  def __init__(self):
    session.Session.global_vars["TIMESTAMP"] = TimeStampBuiltin()
    session.Session.global_vars["LOGTIMESTAMP"] = LogTimeStampBuiltin()
    session.Session.global_vars["DATADIR"] = DatadirBuiltin()

    import os
    if "HOME" in os.environ:
      session.Session.global_vars["HOME"] = os.environ["HOME"]

  def clear(self, ses):
    ses._vars = {}

  def addVariable(self, ses, var, expansion):
    ses.setVariable(var, expansion)

  def removeVariables(self, ses, text):
    d = dict(ses._vars)
    d.update(session.Session.global_vars)
    badvariables = utils.expand_text(text, list(d.keys()))
    ret = []
    for mem in badvariables:
      ret.append( (mem, d[mem]) )
      ses.removeVariable(mem)

    return ret

  def defaultResolver(self, args):
    """
    Returns a defaultresolver that will look up potential default values in 
    this session's variables.

    Lookup will be first::

       "default.%s.%s" % (command, argument,)

    then::

       "default.%s" % (argument,)

    to allow for overriding all arguments of a given name, or only arguments 
    for specific commands.

    @returns: a function object that will do the desired lookup
    @rtype: function object
    """
    ses = args["session"]
    command = args["commandname"]

    def resolver(argument, ses=ses, command=command):
      output = ses.getVariable( "default.%s.%s" % (command, argument,))
      if output == None:
        output = ses.getVariable( "default.%s" % (argument,))
      return output
    return resolver

  def expand(self, ses, text):
    t = utils.expand_vars(text, session.Session.global_vars)
    t = utils.expand_vars(t, ses._vars)
    return utils.denest_vars(t, {})

  def expand_command(self, ses, text):
    t = utils.expand_vars(text, session.Session.global_vars)
    return utils.expand_vars(t, ses._vars)

  def getInfo(self, ses, text=""):
    data = list(ses._vars.keys())
    if text:
      data = utils.expand_text(text, data)
    data = ["variable {%s} {%s}" % (m, ses._vars[m]) for m in data]
    return data

  def getItems(self):
    return [ "variable" ]

  def getParameters(self, item):
    if item != "variable":
      raise ValueError("%s is not a valid item for this manager." % item)
    return [ ("var", "The variable"), 
             ("expansion", "The thing the variable expands into." ) ]

  def getInfoMappings(self, item, ses):
    if item != "variable":
      raise ValueError("%s is not a valid item for this manager." % item)
    l = []
    for mem in list(self._vars.keys()):
      l.append( {"var": mem, "expansion": ses._vars[mem]} )

    return l
      
  def getStatus(self, ses):
    return "%d variable(s)." % len(ses._vars)

  def addSession(self, newsession, basesession=None):
    self.clear(newsession)

    if basesession:
      for mem in list(basesession._vars.keys()):
        newsession._vars[mem] = basesession._vars[mem]

  def persist(self, args):
    """
    write_hook function for persisting the state of our session.
    """
    ses = args["session"]
    quiet = args["quiet"]

    data = self.getInfo(ses)
    if quiet == 1:
      data = [m + " quiet={true}" for m in data]

    return data

  def denestVars(self, args):
    """
    Handles denesting variables for Lyntin evaluation mode.
    """
    ses = args["session"]
    internal = args["internal"]
    verbatim = args["verbatim"]
    text = args["dataadj"]

    if verbatim == 1:
      return text

    return utils.denest_vars(text, ses._vars)

  def userfilter(self, args):
    """
    user_filter_hook for handling incoming user data.
    """
    ses = args["session"]
    internal = args["internal"]
    verbatim = args["verbatim"]
    text = args["dataadj"]

    if verbatim == 1:
      return text

    varexpansion = self.expand_command(ses, text)

    if varexpansion == text:
      return text
    else:
      exported.lyntin_command(varexpansion, 1, ses)
      return None


commands_dict = {}

def variable_cmd(ses, args, input):
  """
  Creates a variable for that session of said name with said value.
  Variables can then pretty much be used anywhere.

  examples:
    #variable {hps} {100}
    #action {HP: %0/%1 } {#variable {hps} {%0}}

  Variables can later be accessed via the variable character
  (which defaults to $) and the variable name.  In the case of the
  above, the variable name would be $hps.  

  We also handle braced closures for denoting variables like ${hps}.  
  If you have a variable hps and a variable hpset, you can explicitly
  specify which one using { }.

  There are also system variables $HOME, $TIMESTAMP, $LOGTIMESTAMP,
  and $DATADIR (must be upper-cased) and global variables.  To set 
  a global variable which can be used in all sessions, it must 
  be preceded by a _.

  examples:
    #variable {_fun} {happy fun ball}
    #showme $_fun
    #showme $TIMESTAMP
    #showme ${TIMESTAMP}

  category: commands
  """
  var = args["var"]
  expansion = args["expansion"]
  quiet = args["quiet"]

  vm = exported.get_manager("variable")

  if not expansion:
    data = vm.getInfo(ses, var)
    if not data:
      data = ["variable: no variables defined."]

    exported.write_message("variables:\n" + "\n".join(data), ses)
    return 

  try:
    vm.addVariable(ses, var, expansion)
    if not quiet:
      exported.write_message("variable: {%s}={%s} added." % (var, expansion), ses)

  except Exception as e:
    exported.write_error("variable: cannot be set. %s" % e, ses)

commands_dict["variable"] = (variable_cmd, "var= expansion= quiet:boolean=false")


def unvariable_cmd(ses, args, input):
  """
  Allows you to remove variables.

  category: commands
  """
  func = exported.get_manager("variable").removeVariables
  modutils.unsomething_helper(args, func, ses, "variable", "variables")

commands_dict["unvariable"] = (unvariable_cmd, "str= quiet:boolean=false")

vm = None


def load():
  """ Initializes the module by binding all the commands."""
  global vm
  modutils.load_commands(commands_dict)
  vm = VariableManager()
  exported.add_manager("variable", vm)

  exported.hook_register("user_filter_hook", vm.userfilter, 10)
  exported.hook_register("user_filter_hook", vm.denestVars, 95)
  exported.hook_register("default_resolver_hook", vm.defaultResolver)
  exported.hook_register("write_hook", vm.persist)

def unload():
  """ Unloads the module by calling any unload/unbind functions."""
  global vm
  modutils.unload_commands(list(commands_dict.keys()))
  exported.remove_manager("variable")

  exported.hook_unregister("user_filter_hook", vm.userfilter)
  exported.hook_unregister("user_filter_hook", vm.denestVars)
  exported.hook_unregister("default_resolver_hook", vm.defaultResolver)
  exported.hook_unregister("write_hook", vm.persist)

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#######################################################################
# This file is part of Lyntin.
# copyright (c) Free Software Foundation 2001, 2002
#
# Lyntin is distributed under the GNU General Public License license.  See the
# file LICENSE for distribution details.
# $Id: lyntinunittest.py,v 1.2 2003/11/09 23:25:09 willhelm Exp $
#######################################################################
"""
This module uses the Python unittest framework to unittest various
pieces inside of Lyntin--mostly stuff that's standalone in the 
lyntin.utils module.
"""
# we kind of assume this is being run in ./lyntin40/tools/
import sys, unittest, textwrap
sys.path.insert(0, "../")

import lyntin.utils
import lyntin.ansi

class TestSplitCommands(unittest.TestCase):
  t = (
     ("test", ["test"]),
     ("test;test2", ['test', 'test2']),
     ("#alias t3k #ses a localhost 3000", ["#alias t3k #ses a localhost 3000"]),
     ("#alias gv {put vx;get all}", ["#alias gv {put vx;get all}"]),
     ("#alias sv {put vx;get all};test", ["#alias sv {put vx;get all}", "test"]),
     (r"#sh \{ blah;#sh another }", [r"#sh \{ blah", r"#sh another }"])
  )

  def testSplit(self):
    """Tests lyntin.utils.split_commands"""
    for i in range(len(self.t)):
      c, s = self.t[i]
      result = lyntin.utils.split_commands(";", c)
      self.assertEqual(s, result, f"test {i}")

class TestSplitAnsiFromText(unittest.TestCase):
  t = (
     ( "This is some text.", ["This is some text."]),
     ( "\33[1;37mThis is\33[0m text.", ["\33[1;37m", "This is", "\33[0m", " text."]),
     ( "Hi \33[1;37mThis is\33[0m text.", ["Hi ", "\33[1;37m", "This is", "\33[0m", " text."]),
     ("\33[1;37mThis is\33[0", ["\33[1;37m", "This is", "\33[0"])
  )

  def testAnsiSplit(self):
    """Tests lyntin.ansi.split_ansi_from_text"""
    for i in range(len(self.t)):
      c, s = self.t[i]
      result = lyntin.ansi.split_ansi_from_text(c)
      self.assertEqual(s, result, f"test {i}")


class TestWrapText(unittest.TestCase):
  text = "This is a really long line to see if we're wrapping correctly.  Because it's way cool when we write code that works.  Yay!"
  text2 = "Hi.  Check this out: Thistexthasnospacesinitandmightcausethingstocrashorgointoaninfiniteloopandstuff.whichwouldbesuperbad.  What do you think?"
  text3 = "[You said to notadragon17]: http://abcnews.go.com/sections/us/DailyNews/terrifica021105.html"
  text4 = "[notadragon17 said]: 123456789012345678901234567890124567890123456789012345678901234567890123456789012345678901234567890"
  text5 = "This is some text \33[1;37mwith some\33[0m ansi formatting in it to see if we can handle wrapping with it \33[1;37mtoo.\33[0m"

  def wrap_text(text, indent=0, firstline=0):
    """
    Wraps the text into lines with an optional indent and custom first-line indent.
    """
    wrapper = textwrap.TextWrapper(width=50, initial_indent=" " * indent, subsequent_indent=" " * indent)
    lines = wrapper.wrap(text)
    if firstline:
      lines[0] = " " * indent + lines[0]
    return "\n".join(lines)

  def testWrapText2(self):
    """Tests lyntin.ansi.wrap_text with a long word in it"""
    self.assertEqual(lyntin.utils.wrap_text(self.text2), \
"""Hi.  Check this out: 
Thistexthasnospacesinitandmightcausethingstocrash
orgointoaninfiniteloopandstuff.whichwouldbesuperb
ad.  What do you think?""", "test 1")

    self.assertEqual(lyntin.utils.wrap_text(self.text2, indent=5), \
"""Hi.  Check this out: 
     Thistexthasnospacesinitandmightcausethingsto
     crashorgointoaninfiniteloopandstuff.whichwou
     ldbesuperbad.  What do you think?""", "test 2")

    def testWrapText2(self):
      """Tests lyntin.ansi.wrap_text with a long word in it"""
      self.assertEqual(
        lyntin.utils.wrap_text(self.text2),
        """Hi.  Check this out: 
Thistexthasnospacesinitandmightcausethingstocrash
orgointoaninfiniteloopandstuff.whichwouldbesuperb
ad.  What do you think?""",
        "test 1"
      )

      self.assertEqual(
        lyntin.utils.wrap_text(self.text2, indent=5),
        """Hi.  Check this out: 
     Thistexthasnospacesinitandmightcausethingsto
     crashorgointoaninfiniteloopandstuff.whichwou
     ldbesuperbad.  What do you think?""",
        "test 2"
      )

  def testWrapText3(self):
    """Tests lyntin.ansi.wrap_text with a long url in it"""
    self.assertEqual(lyntin.utils.wrap_text(self.text3, 70, 5, 0), \
"""[You said to notadragon17]: 
     http://abcnews.go.com/sections/us/DailyNews/terrifica021105.html
     """, "test 1")

  def testWrapText4(self):
    """Tests lyntin.ansi.wrap_text with a long sequence of digits in it"""
    self.assertEqual(lyntin.utils.wrap_text(self.text4, 70, 5, 0), \
"""[notadragon17 said]: 
     1234567890123456789012345678901245678901234567890123456789012345
     67890123456789012345678901234567890""", "test 1")

  def testWrapText5(self):
    """Tests lyntin.ansi.wrap_text with a ansi markup in it"""
    self.assertEqual(lyntin.utils.wrap_text(self.text5), \
"""This is some text \33[1;37mwith some\33[0m ansi formatting in 
it to see if we can handle wrapping with it \33[1;37mtoo.\33[0m""", "test 1")

    self.assertEqual(lyntin.utils.wrap_text(self.text5, indent=5), \
"""This is some text \33[1;37mwith some\33[0m ansi formatting in 
     it to see if we can handle wrapping with 
     it \33[1;37mtoo.\33[0m""", "test 2")

class TestParseTimespan(unittest.TestCase):
  t = (
    ("1h", 3600),
    ("1m", 60),
    ("1s", 1),
    ("1h2m3s", 3723),
    ("17", 17),
    ("5h", 3600 * 5)
  )

  def testParseTimespam(self):
    """Tests lyntin.utils.parse_timespam"""
    for i in range(len(self.t)):
      c, s = self.t[i]
      self.assertEqual(lyntin.utils.parse_timespan(c), s, f"test {i}")


class TestExpandPlacementVars(unittest.TestCase):
  c = "#test 1 2 3"
  t = (
    ("#test", "#test 1 2 3"),
    ("#test %1 %2", "#test 1 2"),
    ("#test %0", "#test #test"),
    ("#test %-1", "#test 3"),
    ("#test %:-1", "#test #test 1 2"),
    ("#test %1:-1", "#test 1 2")
  )

  def testExpandPlacementVars(self):
    """tests lyntin.utils.expand_placement_vars"""
    from lyntin.utils import expand_placement_vars
    for i in range(0, len(self.t)):
      c, s = self.t[i]
      self.assertEqual(expand_placement_vars(self.c, c), s, "test %d" % i)

class TestExpandVars(unittest.TestCase):
  varmap = {"var1": "value1", "var2": "value2", "var3": "value3"}
  t = (
    (r"$", "$"),
    (r" $", " $"),
    (r"$ ", "$ "),
    (r"This has no vars.", "This has no vars."),
    (r"$var1 $var2 $var3", "value1 value2 value3"),
    (r"$var1 $$var2 \$var3", r"value1 $$var2 \$var3"),
    (r"${var1} $${var2} \${var3}", r"value1 $${var2} \${var3}")
  )

  def testExpandVars(self):
    """tests lyntin.utils.expand_vars"""
    from lyntin.utils import expand_vars
    for i in range(len(self.t)):
      c, s = self.t[i]
      self.assertEqual(expand_vars(c, self.varmap), s, f"test {i}")


"""
# FIXME - these always fail because we don't get the precision right.
# not sure what to do about that.
from lyntin.utils import parse_time
_pass_fail("parse_time 1", int(parse_time("4:20p")), 1029878400)
_pass_fail("parse_time 2", int(parse_time("4m")), 1029796956)
_pass_fail("parse_time 3", int(parse_time("9")), 1029796725)
_pass_fail("parse_time 4", int(parse_time("1:17:34a")), 1029824254)
"""


if __name__ == '__main__':
  from lyntin import constants
  print((constants.VERSION))

  unittest.main(testRunner=unittest.TextTestRunner(verbosity=2))

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: manager.py,v 1.5 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
Managers manage things.  Most X{manager}s subclass the "manager.Manager"
class.  It doesn't really provide a lot of functionality, but it 
allows us to group them all and treat them all the same.  Adding 
new managers is much easier because of this.

Also, managers register themselves with the engine via the
"exported.add_manager" function.  The engine will cycle
through registered managers for things like status and persistence.
In addition, registered managers get told when the user has created
a new session and when the user has ended a session through the
"addSession" and "removeSession" methods.

To build a new manager, you need to:

  1. extend the manager.Manager class

  2. implement all the methods of manager.Manager that are marked as 
     needing to be overridden

  3. implement the additional methods that your manager needs

  4. create a "load()" function in the module your manager is 
     defined in which adds the manager to the engine via 
     "exported.add_manager(...)"


Then to dynamically load your new module and instantiate your new manager
you can do one of two things:

  1. put the .py file in the modules/ subdirectory where it will be 
     loaded automatcially

  2. execute an "#import modulenamehere" inside of Lyntin which will 
     import the module
"""

class Manager:
  """
  Base manager class for managing things in Lyntin.  The Manager class
  gives all managers a standard way of interacting with sessions and user
  queries for information, 
  """
  def __init__(self):
    pass

  def clear(self, ses=None):
    """
    Removes everything the manager was managing--essentially reinitializes it.
    Override this to clear out the data your manager is managing.  This is
    typically session oriented and gets called by the "#clear" command.

    @param ses: the session this applies to--None if it applies to all
        sessions
    @type  ses: session.Session
    """
    pass

  def getInfo(self, ses, text=''):
    """
    Returns information managed by this class.  This is mostly for 
    display to the user--we shouldn't be using this method for Lyntin 
    introspection.

    @param text: allows the user to pass in a name filter which should
        show a subset of the data based on what names match the filter
    @type  text: string

    @return: a string of everything involved
    @rtype: string
    """
    return ''

  def getItems(self):
    """
    Returns a list of the items that this manager manages.  So the
    gag manager manages gags as well as antigags and would return::

       [ "gag", "antigag" ]

    These items are used in getInfoMappings as well as getParameters.

    If this manager doesn't manage anything, then it'll return an
    empty list.

    @return: returns a list of strings
    @rtype: list of strings
    """
    return []

  def getInfoMappings(self, item, ses):
    """
    Returns a list of maps of parameter name -> value that represents
    all the info this manager is managing for this session.

    For example, an AliasManager manages aliases and their expansions.
    Say it had three aliases a, b, and c which expand to "smile %1", 
    "frown %1", and "kick %1".  It would return::

       [
         { "alias": "a", "expansion": "smile %1" },
         { "alias": "b", "expansion": "frown %1" },
         { "alias": "c", "expansion": "kick %1" }
       ]

    @param item: the item to get data for
    @type  item: string

    @param ses: the session in question
    @type  ses: Session

    @returns: a list of maps of parameter name -> value
    @rtype: list of mappings
    """
    return []

  def getParameters(self, item):
    """
    Returns a list of tuples of the parameters we're storing in
    this manager and the description of each parameter.

    If this manager does not manage this item, it should raise
    a ValueError.

    @param item: the item to get parameters for
    @type  item: string

    @returns: list of (parameter, desc) tuples
    @rtype: list of tuples
    """
    return []

  def addSession(self, newsession, basesession=None):
    """
    Tells the manager to create a new session based on another session.
    For example, when we connected to the 3k mud, we would tell all
    the managers to clone the common session to the new session created
    thus populating the new session.

    @param newsession: the new session just created
    @type  newsession: Session

    @param basesession: the session to use as a template for the new
        session.  most managers just copy the data from the basesession
        to the newsession.  if this is None, then we don't want to clone
        from anything--we use None when Lyntin starts up and we create
        the common session.
    @type  basesession: Session
    """
    pass

  def removeSession(self, ses):
    """
    Tells the manager to remove information regarding the session.

    @param ses: the session we no longer need to hold information for
        because the session has gone away
    @type  ses: Session
    """
    pass

  def getStatus(self, ses):
    """
    Returns a one-liner status of the state of this manager for
    a given session.  If this manager does not apply to sessions
    then it should return an empty string.

    For example, the SubstituteManager which is holding 5 substitutes
    and 2 gags for a session named "3k" would return the string::

       "5 substitute(s). 2 gag(s)."

    But the ThreadManager which is a globally scoped manager and doesn't 
    apply to the "3k" session would return an empty string.

    @param ses: the session to get status for
    @type  ses: Session

    @return: a one-liner string of the status or an empty string
    @rtype: string
    """
    return ''

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#!/usr/bin/env python
#######################################################################
# This file is part of Lyntin.
# copyright (c) Free Software Foundation 2001
#
# Lyntin is distributed under the GNU General Public License license.  See the
# file LICENSE for distribution details.
# $Id: connection.py,v 1.2 2005/01/07 14:50:40 glasssnake Exp $
#######################################################################
"""
This new test-server is a patchwork of stuff from the existing test server
and code I wrote for the Varium mud server way back when.  It is actually
a functional mini-mud now.
"""
import testserver, toolsutils
from toolsutils import color, wrap_text


class Connection:
    def __init__(self, world, newsock, newaddr=''):
        self._world = world
        self._sock = newsock
        self._sock.setblocking(0)  # non-blocking
        self._buffer = []
        self._addr = newaddr
        self._name = "spirit"
        self._desc = "A regular user."

        self._dir = [item for item in dir(self.__class__) if
                     item.startswith("handle_") and callable(getattr(self, item))]

    def __str__(self):
        return repr(self._addr)

    def killConn(self):
        """Shuts down the socket for the Connection object."""
        if not self._sock:
            return

        try:
            print(f"DEBUG: Closing down socket: {self._sock.fileno()} ")
            self._sock.shutdown(2)
            self._sock.close()
        except OSError as e:
            print(f"DEBUG: Error shutting down socket: {e}")

        self._sock = None
        if self in self._world._ms._conns:
            self._world._ms._conns.remove(self)  # Remove from connection list safely
        self._world.disconnect(self)

    def write(self, data):
        """
        Sends a string or byte data to the socket.
        """
        if not self._sock:
            return

        # Check if socket is already closed
        if self._sock.fileno() == -1:
            print("DEBUG: Checking is socket is closed before write()")
            self._sock = None  # Ensure cleanup
            return

        if not data:
            print("DEBUG: write() failed - Data is empty")
            return

        # Ensure data is bytes before sending
        if isinstance(data, str):
            data = data.replace("\n", "\r\n")  # Convert \n to \r\n for telnet/MUD clients
            data = data.encode('utf-8')

        try:
            self._sock.send(data)
        except (BrokenPipeError, ConnectionResetError):
            self._sock.close()
            self._sock = None
        except OSError as e:
            self._sock.close()
            self._sock = None
        except Exception as e:
            print(f"DEBUG: Unexpected error: {e}. Closing socket...")
            self._sock.close()
            self._sock = None

    def color(self, data, fc=37, bc=40, bold=0):
        """Formats text with ANSI escape codes."""
        escape = "\033"
        if bold == 1:
            return f"{escape}[1;{fc};{bc}m{data}{escape}[0m"
        else:
            return f"{escape}[{fc};{bc}m{data}{escape}[0m"

    def sockid(self):
        if self._sock:
            return self._sock    # Explicitly return None if the socket is invalid

    def handleNetworkData(self, new_data):
        for c in new_data:
            if c in (chr(8), chr(127)):
                if self._buffer:
                    self._buffer = self._buffer[:-1]
                continue
            elif c == "\n":
                self._world.enqueue(testserver.InputEvent(self, "".join(self._buffer)))
                self._buffer = []
                continue
            elif c in ("\r", chr(0)):
                pass
            else:
                self._buffer.append(c)

    def handleInput(self, world, input):
        comm = input.split(" ", 1)[0]

        # Look for the method in self
        method_name = "handle_" + comm
        if hasattr(self, method_name):
            method = getattr(self, method_name)
            method(world, input)
        else:
            # Handle invalid command
            self.write(color("huh? '%s'\n" % input, 35))

        # Always display prompt
        self.write("> ")

    def handle_set(self, world, user_input):
        """ Lets you set things: (name, desc...)."""
        if " " in user_input:
            user_input = user_input.split(" ", 1)[1]
        if " " in user_input:
            name, value = user_input.split(" ", 1)

            if hasattr(self, '_%s' % name):
                self.write("Old value of %s is '%s'.\n" % (name, eval("self._%s" % name)))
            exec("self._%s = value" % name)
            self.write("Set to '%s'.\n" % value)
        else:
            self.write("Nothing to set.\n")

    def handle_say(self, world, user_input):
        """ Talk to your fellow mudders!"""
        if " " in user_input:
            text = user_input.split(" ", 1)[1]
            self.write(toolsutils.wrap_text("You say: %s" % text, 72, 5, 0) + "\n")
            world.spamroom(toolsutils.wrap_text("%s says: %s" % (self._name, text), 72, 5, 0) + "\n")
        else:
            self.write("say what?\n")

    def handle_look(self, world, user_input):
        """ Lets you look at things.  syntax: look <at thing>"""
        item = None
        if " " in user_input:
            item = user_input.split(" ", 1)[1].replace("at", "").strip().lower()

        self.write(world.look(self, item))

    def handle_quit(self, world, user_input):
        """ Quits out."""
        self.killConn()

    def handle_help(self, world, text):
        """ Prints out all the commands we understand."""
        commands = []
        for mem in self._dir:
            if mem.find("handle_") == 0:
                doc = ""
                try:
                    doc = eval("self.%s.__doc__" % mem)
                except:
                    pass

                if doc:
                    commands.append(mem[7:] + " - " + doc)
                else:
                    commands.append(mem[7:])

        self.write("\n".join(commands) + "\n")

    def handle_set_color(self, world, text):
        """ Sets the terminal text color to yellow. """
        try:
            # Write the yellow text
            self.write("\033[33m\nNow yellow.")  # \033[0m resets the color
        except Exception as e:
            # Handle exceptions and provide feedback
            self.write(f"Error setting color: {e}\n")

    def handle_colors(self, world, text):
        """ Prints out all the colors we know about. """
        try:
            response = []

            # Iterate through the range of background colors (40 to 47)
            for background in range(40, 48):
                # Iterate through the range of foreground colors (30 to 37)
                for foreground in range(30, 38):
                    # Add normal color
                    response.append(self.color(str(foreground), foreground, background))
                    # Add bold color
                    response.append(self.color(str(foreground), foreground, background, bold=1))

                # Reset colors and add a newline
                response.append("\033[0m\n")

            # Join all parts of the response and write it out
            self.write(''.join(response))

        except Exception as e:
            # Log error message if any issue occurs
            self.write(f"Error while handling colors: {e}")

    def handle_lyntin(self, world, text):
        """ Returns a paragraph, which coincidentally, is a description of Lyntin."""
        output = ("Lyntin is a mud client that is written in Python and uses " +
                  "Python as a scripting language. It strives to be functionally " +
                  "similar to TinTin++ while enhancing that functionality with " +
                  "the ability to call Python functions directly from the input " +
                  "line. It has the advantage of being platform-independent and " +
                  "has multiple interfaces as well--I use Lyntin at home with " +
                  "the Tk interface as well as over telnet using the text " +
                  "interface.\n")
        output = toolsutils.wrap_text(output, 70, 0, 0)
        self.write(output)

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#######################################################################
# This file is part of Lyntin.
# copyright (c) Free Software Foundation 2001-2007
#
# Lyntin is distributed under the GNU General Public License license.  See the
# file LICENSE for distribution details.
# $Id: message.py,v 1.3 2007/07/24 00:39:03 willhelm Exp $
#######################################################################
"""
Holds the ui's Message class.  This gets passed around Lyntin and
allows us to scope data going to the ui.
"""

""" The message type constants."""
ERROR = "ERROR: "
USERDATA = "USERDATA: "
MUDDATA = "MUDDATA: "
LTDATA = "LTDATA: "

""" Used for debugging purposes."""
MESSAGETYPES = {ERROR: "ERROR: ",
                USERDATA: "USERDATA: ",
                MUDDATA: "MUDDATA: ",
                LTDATA: "LTDATA: "}

class Message:
  """
  Encapsulates a message to be written to the user.
  """
  def __init__(self, data, messagetype=LTDATA, ses=None, **hints):
    """
    Initialize.

    @param data: the message string
    @type  data: string

    @param messagetype: the message type (use a constant defined in ui.ui)
    @type  messagetype: int

    @param ses: the session this message belongs to
    @type  ses: session.Session
    """
    self.session = ses
    self.data = data
    self.type = messagetype
    self.hints = hints

  def __repr__(self):
    """
    Represents the message (returns data + type).
    """
    return repr(self.session) + MESSAGETYPES[self.type] + repr(self.data)

  def __str__(self):
    """
    The string representation of the Message is the data
    itself.
    """
    return self.data





#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: scheduler.py,v 1.5 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This module defines the ScheduleManager which manages scheduling 
events for Lyntin.  It's pretty intense.  It handles both events
oriented around "lyntin time" (what the current Lyntin tick is)
as well as "real time" (what time it really is).  Scheduled
events can be lyntin commands as well as functions with 
arguments.

This module implements the #schedule and #unschedule commands
as well as the completely re-implemented #tick* suite of
commands.
"""
import time
from lyntin import exported, manager, utils, event
from lyntin.modules import modutils

myscheduler = None
commands_dict = {}

def truncate(text, maxwidth):
  if len(text) > maxwidth:
    text = text[:maxwidth] + "..."
  return text

class SchedEvent:
  """ 
  Holds event data as well as handles representation of the data.
  """
  def __init__(self, offset, ses, cmd, repeat=0, quiet=0, tag="none"):
    self._id = "-1"
    self._tag = tag
    self._offset = offset
    self._ses = ses
    self._cmd = cmd
    self._repeat = repeat
    self._quiet = quiet
    self._args = []
    self._xargs = {}

  def __repr__(self):
    if self._repeat == 0:
      return truncate("%s [%s] %d {%s}" % (self._id, self._ses._name, self._offset, self._cmd), 60)
    return truncate("%s [%s] %d(r) {%s}" % (self._id, self._ses._name, self._offset, self._cmd), 60)

class SchedTimeEvent:
  """
  Holds event data as well as handles representation of the data.

  The only difference between this and the SchedEvent is that this
  displays the offset as if it were a real date/time (which it is).
  """
  def __init__(self, offset, ses, cmd, repeat=0, quiet=0, tag="none"):
    self._id = "-1"
    self._tag = tag
    self._offset = offset
    self._ses = ses
    self._cmd = cmd
    self._repeat = repeat
    self._quiet = quiet
    self._args = []
    self._xargs = {}
    self._time = ""

  def __repr__(self):
    return truncate("%s [%s] %s {%s}" % (self._id, self._ses._name, time.strftime("%d %b %H:%M:%S", time.localtime(self._offset)), self._cmd), 60)



class Scheduler:
  """
  Manages scheduled data.
  """
  def __init__(self):
    # mapping of tick time -> SchedEvent list
    self._events = {}

    # mapping of seconds since epoch -> SchedEvent list
    self._tevents = {}

    # the current Lyntin tick
    self._current_tick = 0

    # the event id index
    self._eid = 0

  def startup(self):
    exported.hook_register("timer_hook", self.timeUpdate)

  def shutdown(self):
    exported.hook_unregister("timer_hook", self.timeUpdate)
 
  def getEvents(self, ses):
    """
    Returns a list of the events for a given session.

    @param ses: the session to look at
    @type  ses: Session

    @returns: the list of session information
    @rtype: list of strings
    """
    output = []
    for listing in list(self._events.values()):
      for mem in listing:
        if mem._ses == ses:
          output.append(repr(mem))

    for listing in list(self._tevents.values()):
      for mem in listing:
        if mem._ses == ses:
          output.append(repr(mem))

    output.sort()
    return output

  def getEventById(self, id):
    """
    Finds an event by id or by tag.  It returns the event as well
    as adding a _next_tick attribute to the event telling you
    when the event is next scheduled to execute.  Sneaky, eh?

    @param id: the id or tag of the event to find
    @type  id: string

    @returns: the SchedEvent instance or None
    @rtype: SchedEvent
    """
    for key in list(self._events.keys()):
      listing = self._events[key]
      for mem in listing:
        if mem._id == id or mem._tag == id:
          mem._next_tick = key
          return mem

    for key in list(self._tevents.keys()):
      listing = self._tevents[key]
      for mem in listing:
        if mem._id == id or mem._tag == id:
          mem._next_tick = key
          return mem

    return []

  def removeById(self, id):
    """
    Removes an event from the scheduler by id or by tag or by '*'.

    @param id: the id or tag of the event to remove
    @type  id: string

    @returns: a list of the events unscheduled
    @rtype: list of strings
    """
    output = []
    for listing in list(self._events.values()):
      for mem in listing:
        if id == '*' or mem._id == id or mem._tag == id:
          output.append(repr(mem))
          listing.remove(mem)

    for listing in list(self._tevents.values()):
      for mem in listing:
        if id == '*' or mem._id == id or mem._tag == id:
          output.append(repr(mem))
          listing.remove(mem)

    output.sort()
    return output

  def addEvent(self, tick, sevent, real=0, id=-1):
    """
    Adds an event to the scheduler.

    @param tick: the tick that this event kicks off at
    @type  tick: int

    @param sevent: the SchedEvent object
    @type  sevent: SchedEvent

    @param real: whether or not this is a real event (tick is seconds since
        epoch) or a regular event (tick is an offset of seconds from now)
    @type  real: int
    """
    if id == -1:
      eid = self._eid
      self._eid += 1
    else:
      eid = id

    sevent._id = str(eid)

    if real == 0:
      eventdict = self._events
    else:
      eventdict = self._tevents

    if tick in eventdict:
      eventdict[tick].append(sevent)
    else:
      eventdict[tick] = [sevent]

  def timeUpdate(self, args):
    """
    This gets called by the timer_hook in the engine every
    second.  It goes through and executes all the events for this
    Lyntin tick as well as events who are supposed to execute at
    this seconds since the epoch or before.

    It also handles tossing events back in the schedule if they
    need repeating.
    """
    tick = args["tick"]

    events = []

    if tick in self._events:
      for mem in self._events[tick]:
        events.append(mem)

      del self._events[tick]

    self._current_tick = tick

    # we want to execute for this second and any previous seconds
    # that have been missed.
    sec = int(time.time())
    keys = list(self._tevents.keys())
    keys = [mem for mem in keys if mem < sec]
    for key in keys:
      for mem in self._tevents[key]:
        events.append(mem)
      del self._tevents[key]

    # go through and execute all the events we've found
    for mem in events:
      if not mem._quiet:
        exported.write_message("Executing %r." % mem)

      if not callable(mem._cmd):
        exported.lyntin_command(mem._cmd, internal=1, session=mem._ses)
      else:
        try:
          mem._cmd(*mem._args, **mem._xargs)
        except:
          exported.write_traceback("exception kicked up while trying to execute event.")

      # handles repeating events
      if mem._repeat == 1:
        self.addEvent(tick + mem._offset, mem, id=mem._id)

def schedule_cmd(ses, args, input):
  """
  With no arguments lets you view the scheduled events.

    lyntin: Scheduled events:
    lyntin: 1 [a] 200 {#showme Will is super duper!}

  First column is the event id.
  Second column is the session it's in.
  Third column is the tick offset or time it's going to kick off at.
  Fourth column is the command to execute.

  With arguments it creates a scheduled event to kick off (and 
  possibly repeat) at TICK seconds from now at which point it will 
  execute EVENT which could be any valid user input.

  examples:

    #schedule {5} {#showme blah}

  will kick off 5 ticks from now (a tick is approx one second) and
  will execute "#showme blah".

    #schedule {1m30s} {#showme blah}

  will kick off in 1 minute and 30 seconds.

    #schedule {10} {#showme blah} {true}

  will kick off every 10 seconds.

  category: commands
  """
  global myscheduler

  tick = args["tick"]
  cmd = args["event"]
  quiet = args["quiet"]
  repeat = args["repeat"]

  if not tick:
    output = myscheduler.getEvents(ses)
    if not output:
      exported.write_message("schedule: there are no scheduled events.")
      return

    if not quiet:
      exported.write_message("scheduled events:\n" + "\n".join(output))
    return

  setimespan = 0
  setime = 0
  try:
    setimespan = utils.parse_timespan(tick)
  except:
    try:
      setime = utils.parse_time(tick)
    except:
      exported.write_error("schedule: %s is not a valid time or timespan." % tick)
      return

  if setimespan != 0:
    sevent = SchedEvent(setimespan, ses, cmd, repeat, quiet)
    tick = setimespan + myscheduler._current_tick
    myscheduler.addEvent(tick, sevent)

  else:
    repeat = 0
    sevent = SchedTimeEvent(setime, ses, cmd, repeat, quiet)
    myscheduler.addEvent(setime, sevent, real=1)

  if not quiet:
    exported.write_message("schedule: event scheduled: %r" % sevent)

commands_dict["schedule"] = (schedule_cmd, "tick= event= repeat:boolean=false quiet:boolean=false")


def unschedule_cmd(ses, args, input):
  """
  Allows you to remove a scheduled event by id.  To remove all events
  scheduled use *.  To see a list of the events and ids for the current 
  session use the #sched command.

  examples:
    #unschedule *
    #unschedule 44

  category: commands
  """
  global myscheduler

  id = args["str"]
  quiet = args["quiet"]

  if id:
    ret = myscheduler.removeById(id)
    if not ret:
      if id == "*":
        exported.write_error("unschedule: no scheduled events to unschedule.")
      else:
        exported.write_error("unschedule: id '%s' is not valid." % id)
      return
    if not quiet:
      exported.write_message("events removed:\n%s" % "\n".join(ret))
    return
 
  exported.write_message("not implemented yet.")

commands_dict["unschedule"] = (unschedule_cmd, "str= quiet:boolean=false")


# this is a dict holding the DEFAULT_TICKER values.  we copy
# this whenever we don't have them.
DEFAULT_TICKER = {
      "len": 3,
      "warn_len": 2,
      "enabled": 0
   }

def _tickfunc(ses):
  """
  Handles executing the command or displaying a message to the
  user.

  @param ses: the Session instance
  @type  ses: Session
  """
  am = exported.get_manager("alias")
  if am:
    tickaction = am.getAlias(ses, "TICK!!!")
  if not tickaction:
    tickaction = am.getAlias(ses, "TICK")

  if tickaction:
    event.InputEvent(tickaction, internal=1, ses=ses).enqueue()
  else:
    exported.write_message("TICK!!!")

def _tickwarnfunc(ses, warnlen): 
  """
  Handles executing the command or displaying a message to the
  user.

  @param ses: the Session instance
  @type  ses: Session

  @param warnlen: the warning length
  @type  warnlen: int
  """
  am = exported.get_manager("alias")
  if am:
    tickaction = am.getAlias(ses, "TICKWARN!!!")
  if not tickaction:
    tickaction = am.getAlias(ses, "TICKWARN")

  if tickaction:
    event.InputEvent(tickaction, internal=1, ses=ses).enqueue()
  else:
    exported.write_message("ticker: %d seconds to tick!" % warnlen)


def _addtickevents(ses):
  """
  Utility function for adding the tick and tickwarn events to 
  the schedule.

  @param ses: the Session instance
  @type  ses: Session
  """
  global myscheduler

  tick_tagname = ses.getName() + "tick"
  tickwarn_tagname = ses.getName() + "tickwarn"

  # build the tick event, figure out when it should start,
  # and add it to the schedule
  sevent = SchedEvent(ses._ticker["len"], ses, _tickfunc, repeat=1, 
                      quiet=1, tag=tick_tagname)
  sevent._args = [ses]
  tick = myscheduler._current_tick + ses._ticker["len"]
  myscheduler.addEvent(tick, sevent)

  # build the tickwarn event, figure out when it should start,
  # and add it to the schedule
  warnsevent = SchedEvent(ses._ticker["len"], ses, _tickwarnfunc, 
                          repeat=1, quiet=1, tag=tickwarn_tagname)
  warnsevent._args = [ses, ses._ticker["warn_len"]]
  tick = tick - ses._ticker["warn_len"]
  myscheduler.addEvent(tick, warnsevent)


def _removetickevents(ses):
  """
  Utility function for removing the tick events (tick and tickwarn)
  for a given session.  We return all the events that we've removed
  so the caller knows whether there were events to remove or not.

  @param ses: the Session instance
  @type  ses: Session

  @returns: the SchedEvents removed
  @rtype: list of SchedEvents
  """
  global myscheduler

  sesname = ses.getName()

  # remove the tick and tickwarn events by id
  tick_tagname = sesname + "tick"
  tickwarn_tagname = sesname + "tickwarn"

  ret = myscheduler.removeById(tick_tagname)
  ret2 = myscheduler.removeById(tickwarn_tagname)
  return ret + ret2
 
def tick_cmd(ses, args, input):
  """
  Displays the number of seconds left before this session's
  ticker ticks.

  When a tick happens, it will look for a TICK!!! alias then a TICK
  alias.  Finding none, it will print TICK!!! to the ui.

  When a tickwarning happens, it will look for a TICKWARN!!! alias
  and then a TICKWARN alias.  Finding none, it will print a tickwarning
  message to the ui.

  This allows you to perform an event every x number of seconds.

  see also: tick, tickon, tickoff, ticksize, tickwarnsize

  category: commands
  """
  global myscheduler
  if not hasattr(ses, "_ticker"):
    ses._ticker = DEFAULT_TICKER.copy()

  if ses._ticker["enabled"] == 1:
    tick = myscheduler._current_tick
    sevent = myscheduler.getEventById(ses.getName() + "tick")
    delta = sevent._next_tick - tick

    exported.write_message("tick: next tick in %d seconds." % delta, ses)
  else:
    exported.write_message("tick: ticker is not enabled.", ses)

commands_dict["tick"] = (tick_cmd, "")


def tickon_cmd(ses, args, input):
  """
  Turns on the ticker for this session.

  see also: tick, tickon, tickoff, ticksize, tickwarnsize

  category: commands
  """
  global myscheduler
  if not hasattr(ses, "_ticker"):
    ses._ticker = DEFAULT_TICKER.copy()

  tick_tagname = ses.getName() + "tick"
  tickwarn_tagname = ses.getName() + "tickwarn"

  # quick check to make sure there isn't already a tick event
  # for this session
  if myscheduler.getEventById(tick_tagname):
    exported.write_error("tickon: ticker is already enabled.", ses)
    return

  _addtickevents(ses)
  ses._ticker["enabled"] = 1

  exported.write_message("tickon: session %s ticker enabled." % ses.getName(), ses)
 
commands_dict["tickon"] = (tickon_cmd, "")


def tickoff_cmd(ses, args, input):
  """
  Turns off the ticker for this session.

  see also: tick, tickon, tickoff, ticksize, tickwarnsize

  category: commands
  """
  _removetickevents(ses)
  ses._ticker["enabled"] = 0

  exported.write_message("tickoff: session %s ticker disabled." 
                         % ses.getName(), ses)

commands_dict["tickoff"] = (tickoff_cmd, "")


def ticksize_cmd(ses, args, input):
  """
  Sets and displays the number of seconds between ticks for this
  session.

  examples:
    #ticksize
    #ticksize 6
    #ticksize 1h2m30s

  see also: tick, tickon, tickoff, ticksize, tickwarnsize

  category: commands
  """
  if not hasattr(ses, "_ticker"):
    ses._ticker = DEFAULT_TICKER.copy()

  size = args["size"]

  if size == 0:
    exported.write_message("ticksize: ticksize is %d seconds." % 
                           ses._ticker["len"], ses)
    return

  ses._ticker["len"] = size

  ret = _removetickevents(ses)
  if ret:
    _addtickevents(ses)
 
  exported.write_message("ticksize: tick length set to %s." % str(size), ses)

commands_dict["ticksize"] = (ticksize_cmd, "size:timespan=0")


def tickwarnsize_cmd(ses, args, input):
  """
  Sets and displays the number of seconds you get warned before a
  Tick actually happens.

  examples:
    #tickwarnsize
    #tickwarnsize 6
    #tickwarnsize 0

  see also: tick, tickon, tickoff, ticksize, tickwarnsize

  category: commands
  """
  if not hasattr(ses, "_ticker"):
    ses._ticker = DEFAULT_TICKER.copy()

  size = args["size"]

  if size == 0:
    exported.write_message("tickwarnsize: tickwarnsize is %d seconds." % 
                           ses._ticker["warn_len"], ses)
    return

  if size > ses._ticker["len"]:
    exported.write_error("tickwarnsize: tickwarn length cannot be >= " +
                         "to ticksize.\nCurrent ticksize is %s." %
                         ses._ticker["len"], ses)
    return

  ses._ticker["warn_len"] = size

  ret = _removetickevents(ses)
  if ret:
    _addtickevents(ses)
 
  exported.write_message("tickwarnsize: tickwarn length set to %s." %
                         str(size), ses)

commands_dict["tickwarnsize"] = (tickwarnsize_cmd, "size:timespan=0")


def load():
  global myscheduler

  myscheduler = Scheduler()
  myscheduler.startup()
  modutils.load_commands(commands_dict)

def unload():
  global myscheduler

  myscheduler.shutdown()
  myscheduler = None
  modutils.unload_commands(commands_dict)



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: speedwalk.py,v 1.8 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This module defines the speedwalking code.  Speedwalking is highly
configurable and it's actually less like speedwalking in other mudclients
and more like alias shorthand that allows you to quickly do things
a number of times.

First you want to create maps from a character or characters to
an expansion using #swdir.  Then every instance of this character/characters
gets expanded to the expansion.

For example::

  #swdir {n} {north}
  #swdir {s} {south}

"nnsss" will expand to "north;north;south;south;south".

Similarly, "2n3s" will expand to "north;north;south;sout;south".

To handle instances where certain combinations will get expanded, but we
really don't want them to be, we use #swexclude::

  #swexclude {news}
"""

# Originally written 2002 by Sebastian John

import re, string
from lyntin import manager, utils, exported
from lyntin.modules import modutils

class SpeedwalkHash:
  def __init__(self):
    self._dirs = {}
    self.compileRegexp()
    self._excludes = []
  
  def __copy__(self):
    sm = SpeedwalkManager()
    for mem in list(self._dirs.keys()):
      sm.addDir(mem, self._dirs[mem])

    for mem in self._excludes:
      sm.addExclude(mem)
    return sm

  def clearDirs(self):
    """
    Clears all stored speedwalking dirs from the manager.
    """
    self._dirs = {}
    self.compileRegexp()
  
  def addDir(self, alias, dir):
    """
    Adds a speedwalking direction alias to the manager.
    
    @param alias: the speedwalking alias
    @type  alias: string

    @param dir: the expansion for the speedwalking alias
    @type  dir: string

    @raises ValueError: if there exists another alias where the dir is a
        substring.  for example if dir was "n" and there was an alias "ln",
        that would raise a ValueError.
    """
    for mem in list(self._dirs.keys()):
      if mem.find(dir) != -1:
        raise ValueError("possible ambiguity")
    self._dirs[alias] = dir
    self.compileRegexp()
  
  def removeDirs(self, alias):
    """
    Removes the speedwalking alias and only this one (no wildcard patterns
    are possible).
    
    @param alias: speedwalking aliases that match the alias will be removed
    @type  alias: string

    @returns: a list of (alias, dir) tuples of removed speedwalking aliases.
    @rtype: list of (string, string)
    """
    try:
      dir = self._dirs[alias]
      del self._dirs[alias]
    except KeyError:
      return []
    else:
      self.compileRegexp()
      return [(alias, dir)]
  
  def getDirs(self):
    """
    Returns a list of all the speedwalking aliases currently defined.
    
    @returns: the list of (alias, dir) tuples consisting of all the
        speedwalking aliases we're managing
    @rtype: list of (string, string)
    """
    dirs = list(self._dirs.items())
    dirs.sort()
    return dirs
  
  def getDirsInfo(self, text=""):
    """
    Returns information about the speedwalking aliases in here.
    
    This is used by #swdir to tell all the speedwalking aliases involved as
    well as #write which takes this information and dumps it to the file.
    
    @param text: the text to expand on to find aliases that the user is
        interested in
    @type  text: string
    
    @returns: list of strings where each string represents a speedwalk alias
    @rtype: list of strings
    """
    listing = list(self._dirs.keys())
    if text:
      listing = utils.expand_text(text, listing)
    
    listing = ["swdir {%s} {%s}" % (mem, self._dirs[mem]) for mem in listing]
    
    return listing

  def getDirsInfoMappings(self):
    l = []
    for mem in list(self._dirs.keys()):
      l.append( { "alias": mem, "dir": self._dirs[mem] } )

    return l
  
  def getDirStatus(self):
    """
    Returns a one liner about dirs.

    @returns: a one-line about the dirs managed.
    @rtype: string
    """
    return "%d dir(s)." % len(self._dirs)
  
  def compileRegexp(self):
    """
    Compiles the actual speedwalking pattern.
    Also maintains self._aliases the default excludes.
    """
    if self._dirs:
      keys = "|".join(list(self._dirs.keys()))
      regexp = "^(\\d*(%s))+$" % (keys)
      self._regexp = re.compile(regexp)
      self._aliases = list(self._dirs.values())
      self._dirs_available = list(self._dirs.keys())
    else:
      self._regexp = None
      self._aliases = []
      self._dirs_available = []
  
  def clearExcludes(self):
    """
    Clears the list of excludes (things we don't want to expand speedwalking
    on).
    """
    self._excludes = []
  
  def addExclude(self, exclude):
    """
    Adds a speedwalking exclude to the manager.
    
    @param exclude: the exclude to add
    @type  exclude: string
    """
    if exclude not in self._excludes:
      self._excludes.append(exclude)
  
  def removeExcludes(self, exclude):
    """
    Removes a speedwalking exclude (and only one, no wildcards or the like)
    from the manager.
    
    @param exclude: the exclude to remove (we don't accept wildcards here)
    @type  exclude: string

    @returns: list of the excludes removed
    @rtype: list of strings
    """
    badexcludes = utils.expand_text(exclude, self._excludes)

    for mem in badexcludes:
      self._excludes.remove(mem)

    return badexcludes
  
  def getExcludes(self):
    """
    Returns the exclude list we are managing.
    
    @return: the sorted list of excludes being managed
    @rtype: list of strings
    """
    self._excludes.sort()
    return self._excludes
  
  def getExcludesInfo(self, text=""):
    """
    Returns information about the speedwalking excludes in here.
    
    This is used by #swexcl to tell all the excludes involved as well as
    #write which takes this information and dumps it to the file.
    
    @param text: the text to expand on to find excludes that the user
        is interested in
    @type  text: string

    @returns: list of strings where each string represents an exclude
    @rtype: list of strings
    """
    listing = self._excludes
    if text:
      listing = utils.expand_text(text, listing)
    
    listing = ["swexclude {%s}" % mem for mem in listing]
    
    return listing
  
  def getExcludesInfoMappings(self):
    l = []
    for mem in self._excludes:
      l.append( { "exclude": mem } )
    return l

  def getExcludeStatus(self):
    """
    Returns a one-line string describing how many excludes we have.

    @returns: a one liner describing how many excludes we have
    @rtype: string
    """
    return "%d exclude(s)." % len(self._excludes)

  def clear(self):
    """
    Clears both speedwalking dir aliases and excludes.
    """
    self.clearDirs()
    self.clearExcludes()
  

class SpeedwalkManager(manager.Manager):
  def __init__(self):
    self._hashes = {}

  def clearDirs(self, ses):
    if ses in self._hashes:
      self._hashes[ses].clearDirs()

  def addDir(self, ses, alias, dir):
    if ses not in self._hashes:
      self._hashes[ses] = SpeedwalkHash()
    self._hashes[ses].addDir(alias, dir)

  def removeDirs(self, ses, alias):
    if ses in self._hashes:
      return self._hashes[ses].removeDirs(alias)
    return []

  def getDirs(self, ses):
    if ses in self._hashes:
      return self._hashes[ses].getDirs()
    return []

  def getDirsInfo(self, ses, text=""):
    if ses in self._hashes:
      return self._hashes[ses].getDirsInfo(text)
    return []

  def getStatus(self, ses):
    if ses in self._hashes:
      sdata = self._hashes[ses]
      return "%s %s" % (sdata.getDirStatus(), sdata.getExcludeStatus())
    return "0 dir(s). 0 exclude(s)."

  def addExclude(self, ses, exclude):
    if ses not in self._hashes:
      self._hashes[ses] = SpeedwalkHash()
    self._hashes[ses].addExclude(exclude)

  def removeExcludes(self, ses, exclude):
    if ses in self._hashes:
      return self._hashes[ses].removeExcludes(exclude)
    return []

  def getExcludesInfo(self, ses):
    if ses in self._hashes:
      return self._hashes[ses].getExcludesInfo()
    return []

  def getItems(self):
    return ["swdir", "swexclude"]

  def getParameters(self, item):
    if item == "swdir":
      return [ ("alias", "The speedwalk alias."),
               ("dir", "The speedwalk direction expansion.") ]
    if item == "swexclude":
      return [ ("exclude", "The word to exclude from swdir expansion.") ]

    raise ValueError("%s is not a valid item for this manager." % item)

  def getInfoMappints(self, item, ses):
    if item not in ["swexclude", "swdir"]:
      raise ValueError("%s is not a valid item for this manager." % item)

    if ses not in self._hashes:
      return []

    if item == "swdir":
      return self._hashes.getDirsInfoMappings()
    
    return self._hashes.getExcludesInfoMappings()

  def getInfo(self, ses):
    if ses in self._hashes:
      myhash = self._hashes[ses]
      return myhash.getDirsInfo() + myhash.getExcludesInfo() 
    return []

  def clear(self, ses):
    if ses in self._hashes:
      self._hashes[ses].clear()
  
  def addSession(self, newsession, basesession=None):
    if basesession:
      if basesession in self._hashes:
        sdata = self._hashes[basesession]
        for mem in list(sdata._dirs.keys()):
          self.addDir(newsession, mem, sdata._dirs[mem])

        for mem in sdata._excludes:
          self.addExclude(newsession, mem)

  def removeSession(self, ses):
    if ses in self._hashes:
      del self._hashes[ses]

  def persist(self, args):
    """
    write_hook function for persisting the state of our session.
    """
    ses = args["session"]
    quiet = args["quiet"]

    data = self.getInfo(ses)
    if quiet:
      data = [m + " quiet={true}" for m in data]

    return data

  def userfilter(self, args):
    """
    user_filter_hook function to check for speedwalking expansion.
    """
    ses = args["session"]
    internal = args["internal"]
    verbatim = args["verbatim"]
    text = args["dataadj"]
    
    if ses not in self._hashes or exported.get_config("speedwalk", ses) == 0 or verbatim == 1:
      return text

    sdata = self._hashes[ses]

    if not sdata._dirs or not sdata._regexp or text in sdata._excludes \
        or text in sdata._aliases or not sdata._regexp.match(text):
      return text
    
    swdirs = []
    dir = num = ""
    for char in text:
      if char.isdigit():
        num = num + char
      else:
        dir = dir + char
        if dir in sdata._dirs_available:
          if num: count = int(num)
          else: count = 1
          for i in range(count):
            swdirs.append(sdata._dirs[dir])
          dir = num = ""
    
    output = ";".join(swdirs)
    if output == text:
      return text
    else:
      # anything that gets recursed on should be recursed internally
      exported.lyntin_command(output, 1, ses)
      return None


commands_dict = {}

def swdir_cmd(ses, args, input):
  """
  This adds speedwalking aliases and tells you the current speedwalking dirs
  already registered.

  examples:
    #swdir {n} {north}
    #swdir {s} {south}
    #swdir {e} {east}
    #swdir {w} {west}
    #swdir {NE} {northeast}
    #swdir {l} {look}
    ...

  This allows you to string characters together to speedwalk:

    4e2sNE

  which using the above swdirs gets expanded to 
  "east;east;east;east;south;south;northeast" and who wants to type all 
  that?

  see also: swexclude

  category: commands
  """
  # originally written by Sebastian John
  alias = args["alias"]
  dir = args["dir"]
  quiet = args["quiet"]

  # they typed '#swdir dd*' and are looking for matching speedwalking dirs
  if not dir:
    data = exported.get_manager("speedwalk").getDirsInfo(ses, alias)
    if not data:
      data = ["swdir: no speedwalking dirs defined."]

    exported.write_message("swdirs:\n" + "\n".join(data), ses)
    return

  try:
    exported.get_manager("speedwalk").addDir(ses, alias, dir)
    if not quiet:
      exported.write_message("swdir: {%s} {%s} added." % (alias, dir), ses)
  except ValueError as e:
    exported.write_error("swdir: cannot add alias '%s': %s." % (alias, e), ses)

commands_dict["swdir"] = (swdir_cmd, "alias= dir= quiet:boolean=false")


def swexclude_cmd(ses, args, input):
  """
  Adds words that should be excluded from speedwalk expansion as well
  as tells you which words are currently being excluded.

  If you had swdirs "n", "e", "s", and "w", you might want to create
  excludes for the words "sense", "news", "sew", ...  Which are real
  words that you most likely don't want to be expanded.

  examples:
    #swexclude {end}
    #swexclude {news}

  see also: swdir

  category: commands
  """
  # originally written by Sebastian John
  excludes = args["exclude"]
  quiet = args["quiet"]

  # they typed '#swexclude'--print out all current speedwalking excludes
  if len(excludes) == 0:
    data = exported.get_manager("speedwalk").getExcludesInfo(ses)
    if not data:
      data = ["swexcl: no speedwalking excludes defined."]

    exported.write_message("swexcludes:\n" + "\n".join(data), ses)
    return

  for exclude in excludes:
    exported.get_manager("speedwalk").addExclude(ses, exclude)
    if not quiet:
      exported.write_message("swexclude: {%s} added." % exclude, ses)

commands_dict["swexclude"] = (swexclude_cmd, "exclude* quiet:boolean=false")


def unswdir_cmd(ses, args, input):
  """
  Allows you to remove swdirs.

  category: commands
  """
  func = exported.get_manager("speedwalk").removeDirs
  modutils.unsomething_helper(args, func, ses, "swdir", "swdirs")

commands_dict["unswdir"] = (unswdir_cmd, "str= quiet:boolean=false")


def unswexclude_cmd(ses, args, input):
  """
  Allows you to remove swexcludes.

  category: commands
  """
  func = exported.get_manager("speedwalk").removeExcludes
  modutils.unsomething_helper(args, func, ses, "swexclude", "swexcludes")

commands_dict["unswexclude"] = (unswexclude_cmd, "str= quiet:boolean=false")


sm = None

def load():
  """ Initializes the module by binding all the commands."""
  global sm
  modutils.load_commands(commands_dict)
  sm = SpeedwalkManager()
  exported.add_manager("speedwalk", sm)

  exported.hook_register("user_filter_hook", sm.userfilter, 80)
  exported.hook_register("write_hook", sm.persist)

  from lyntin import config
  for mem in exported.get_active_sessions():
    tc = config.BoolConfig("speedwalk", 1, 1,
         "Allows you to turn on and turn off speedwalk handling.")
    exported.add_config("speedwalk", tc, mem)

def unload():
  """ Unloads the module by calling any unload/unbind functions."""
  global sm
  modutils.unload_commands(list(commands_dict.keys()))
  exported.remove_manager("speedwalk")

  exported.hook_unregister("user_filter_hook", sm.userfilter)
  exported.hook_unregister("write_hook", sm.persist)

  # remove configuration items for every session involved
  for mem in exported.get_active_sessions():
    exported.remove_config("speedwalk", mem)

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: substitute.py,v 1.10 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This module defines the SubstituteManager which handles substitutes.
"""
from lyntin import ansi, manager, utils, exported
from lyntin.modules import modutils

class SubstituteData:
  def __init__(self):
    self._substitutes = {}
    self._antisubs = []

  def addSubstitute(self, item, substitute):
    """
    Adds a substitute to the dict.

    @param item: the item to substitute
    @type  item: string

    @param substitute: the thing to substitute in place of "item"
    @type  substitute: string
    """
    self._substitutes[item] = substitute 

  def addAntiSubstitute(self, item):
    """ Adds an antisubstitute."""
    self._antisubs.append(item)

  def clear(self):
    """
    Removes all the substitutes.
    """
    self._substitutes.clear()
    self._antisubs = []

  def removeSubstitutes(self, text):
    """
    Removes substitutes from the list.

    Returns a list of tuples of substitute item/substitute that
    were removed.

    @param text: substitutes matching text will be removed
    @type  text: string

    @returns: list of (item, substitute) tuples of removed substitutes
    @rtype: list of (string, string)
    """
    badsubstitutes = utils.expand_text(text, list(self._substitutes.keys()))

    ret = []
    for mem in badsubstitutes:
      ret.append((mem, self._substitutes[mem]))
      del self._substitutes[mem]

    return ret

  def removeAntiSubstitutes(self, text):
    """
    Removes antisubstitutes from the list.

    @returns: a list of antisubstitutes that were removed.
    @rtype: list of strings
    """
    badsubs = utils.expand_text(text, self._antisubs)

    ret = []
    for mem in badsubs:
      ret.append(mem)
      self._antisubs.remove(mem)

    return ret

  def getSubstitutes(self):
    """
    Returns the keys of the substitute dict.

    @returns: list of all the substitute items
    @rtype: list of strings
    """
    listing = list(self._substitutes.keys())
    listing.sort()
    return listing

  def getAntiSubstitutes(self):
    """
    Returns the list of antisubs that we have.

    @returns: list of all antisubs
    @rtype: list of strings
    """
    return self._antisubs

  def expand(self, text):
    """
    Looks at mud data and performs any substitutes.

    It returns the final text--even if there were no substitutes.

    @param text: the text to expand substitutes in
    @type  text: string

    @return: the (un)adjusted text
    @rtype: string
    """
    # FIXME -- this isn't done correctly.
    if len(text) > 0:
      # check for antisubs first
      for mem in self._antisubs:
        if text.find(mem) != -1:
          return text

      # check for subs
      for mem in list(self._substitutes.keys()):
        text = text.replace(mem, self._substitutes[mem])

    return text 

  def getInfo(self, text=''):
    """
    Returns information about the substitutes in here.

    This is used by #substitute to tell all the substitutes involved
    as well as #write which takes this information and dumps
    it to the file.

    @param text: the text used to figure out which substitutes to provide
        information on
    @type  text: string

    @return: list of strings where each string represents a substitute
    @rtype: list of strings
    """
    listing = list(self._substitutes.keys())
    if text:
      listing = utils.expand_text(text, listing)

    listing = ["substitute {%s} {%s}" % (mem, utils.escape(self._substitutes[mem])) for mem in listing]

    return listing

  def getSubstituteInfoMapping(self):
    l = []
    for mem in list(self._substitutes.keys()):
      l.append( { "item": mem,
                  "substitution": utils.escape(self._substitutes[mem]) } )
    return l

  def getAntiSubstitutesInfo(self, text):
    """
    Returns information about the antisubstitutes in here.

    This is used by #antisubstitute to tell all the antisubstitutes involved
    as well as #write which takes this information and dumps it to the file.

    @param text: the text used to figure out which antisubstitutes to provide
        information on
    @type  text: string

    @return: list of strings where each string represents an antisubstitute
    @rtype: list of strings
    """
    listing = self._antisubs
    if text:
      listing = utils.expand_text(text, listing)

    listing = ["antisubstitute {%s}" % mem for mem in listing]

    return listing

  def getAntiSubstituteInfoMapping(self):
    l = []
    for mem in self._antisubs:
      l.append( {"item": mem } )
    return l

  def getStatus(self):
    """
    Returns a one liner of number of substitutes we're managing.

    @returns: string describing how many substitutes we're managing
    @rtype: string
    """
    subs = len(list(self._substitutes.keys()))
    antisubs = len(self._antisubs)

    return "%d substitute(s).  %d antisub(s)." % (subs, antisubs)


class SubstituteManager(manager.Manager):
  def __init__(self):
    self._subs = {}

  def addSubstitute(self, ses, item, sub):
    if ses not in self._subs:
      self._subs[ses] = SubstituteData()
    self._subs[ses].addSubstitute(item, sub)

  def addAntiSubstitute(self, ses, item):
    if ses not in self._subs:
      self._subs[ses] = SubstituteData()
    self._subs[ses].addAntiSubstitute(item)

  def clear(self, ses):
    if ses in self._subs:
      self._subs[ses].clear()

  def removeSubstitutes(self, ses, text):
    if ses in self._subs:
      return self._subs[ses].removeSubstitutes(text)
    return []

  def removeAntiSubstitutes(self, ses, text):
    if ses in self._subs:
      return self._subs[ses].removeAntiSubstitutes(text)
    return []

  def getSubstitutes(self, ses):
    if ses in self._subs:
      return self._subs[ses].getSubstitutes()
    return []

  def getInfo(self, ses, text=''):
    if ses in self._subs:
      return self._subs[ses].getInfo(text)
    return []

  def getItems(self):
    return [ "substitute", "antisubstitute" ]

  def getParameters(self, item):
    if item == "substitute":
      return [ ("item", "The thing to substitute."),
               ("substitution", "The thing to substitute with.") ]

    if item == "antisubstitute":
      return [ ( "item", "The thing whose presence denotes we shouldn't substitute." ) ]

    raise ValueError("%s is not a valid item for this manager." % item)

  def getInfoMappings(self, item, ses):
    if item not in ["substitute", "antisubstitute"]:
      raise ValueError("%s is not a valid item for this manager." % item)
    
    if ses not in self._subs:
      return []

    if item == "substitute":
      return self._subs[ses].getSubstituteInfoMapping()

    return self._subs[ses].getAntiSubstituteInfoMapping()

  def getAntiSubstitutesInfo(self, ses, text=''):
    if ses in self._subs:
      return self._subs[ses].getAntiSubstitutesInfo(text)
    return []

  def getStatus(self, ses):
    if ses in self._subs:
      return self._subs[ses].getStatus()
    return "0 substitute(s). 0 antisub(s)."

  def addSession(self, newsession, basesession=None):
    if basesession:
      if basesession in self._subs:
        sdata = self._subs[basesession]
        for mem in list(sdata._substitutes.keys()):
          self.addSubstitute(newsession, mem, sdata._substitutes[mem])

  def removeSession(self, ses):
    if ses in self._subs:
      del self._subs[ses]

  def expand(self, ses, text):
    if ses in self._subs:
      return self._subs[ses].expand(text)
    return text

  def persist(self, args):
    """
    write_hook function for persisting the state of our session.
    """
    ses = args["session"]
    quiet = args["quiet"]

    data = self.getInfo(ses) + self.getAntiSubstitutesInfo(ses)
    if not quiet:
      data = [m + " quiet={true}" for m in data]

    return data

  def mudfilter(self, args):
    """
    mud_filter_hook function to perform substitutions on data 
    that comes from the mud.
    """
    ses = args["session"]
    text = args["dataadj"]

    if exported.get_config("ignoresubs", ses, 0) == 0:
      text = self.expand(ses, text)
    return text


commands_dict = {}

def substitute_cmd(ses, args, input):
  """
  With no arguments, prints all substitutes.
  With one argument, prints all substitutes which match the argument.
  Otherwise creates a substitution.

  Braces are advised around both 'item' and 'substitution'.

  category: commands
  """
  item = args["item"]
  substitution = args["substitution"]
  quiet = args["quiet"]

  sm = exported.get_manager("substitute")

  if not substitution:
    data = sm.getInfo(ses, item)
    if not data:
      data = ["substitute: no substitutes defined."]

    exported.write_message("substitutes:\n" + "\n".join(data), ses)
    return 

  sm.addSubstitute(ses, item, substitution)
  if not quiet:
    exported.write_message("substitute: {%s} {%s} added." % (item, substitution), ses)

commands_dict["substitute"] = (substitute_cmd, "item= substitution= quiet:boolean=false")

def unsubstitute_cmd(ses, args, input):
  """
  Allows you to remove substitutes.

  category: commands
  """
  func = exported.get_manager("substitute").removeSubstitutes
  modutils.unsomething_helper(args, func, ses, "substitute", "substitutes")

commands_dict["unsubstitute"] = (unsubstitute_cmd, "str= quiet:boolean=false")

def antisubstitute_cmd(ses, args, input):
  """
  Allows you to create antisubstitutes.

  For any line that contains an antisubstitute, we won't do substitutions
  on it.

  category: commands
  """
  item = args["item"]
  quiet = args["quiet"]

  sm = exported.get_manager("substitute")

  if not item:
    data = sm.getAntiSubstitutesInfo(ses)
    if not data:
      data = ["antisubstitute: no antisubstitutes defined."]

    exported.write_message("antisubstitutes:\n" + "\n".join(data), ses)
    return

  sm.addAntiSubstitute(ses, item)
  if not quiet:
    exported.write_message("antisubstitute: {%s} added." % item, ses)

commands_dict["antisubstitute"] = (antisubstitute_cmd, "item= quiet:boolean=false")

def unantisubstitute_cmd(ses, args, input):
  """
  Allows you to remove antisubstitutes.

  category: commands
  """
  func = exported.get_manager("substitute").removeAntiSubstitutes
  modutils.unsomething_helper(args, func, ses, "antisubstitute", "antisubstitutes")

commands_dict["unantisubstitute"] = (unantisubstitute_cmd, "str= quiet:boolean=false")

sm = None

def load():
  """ Initializes the module by binding all the commands."""
  global sm
  modutils.load_commands(commands_dict)
  sm = SubstituteManager()
  exported.add_manager("substitute", sm)

  exported.hook_register("mud_filter_hook", sm.mudfilter, 50)
  exported.hook_register("write_hook", sm.persist)

  from lyntin import config
  for mem in exported.get_active_sessions():
    tc = config.BoolConfig("ignoresubs", 0, 1,
         "Allows you to turn on and turn off substitutions.")
    exported.add_config("ignoresubs", tc, mem)


def unload():
  """ Unloads the module by calling any unload/unbind functions."""
  global sm
  modutils.unload_commands(list(commands_dict.keys()))
  exported.remove_manager("substitute")

  exported.hook_unregister("mud_filter_hook", sm.mudfilter)
  exported.hook_unregister("write_hook", sm.persist)

  # remove configuration items for every session involved
  for mem in exported.get_active_sessions():
    exported.remove_config("ignoresubs", mem)

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: gag.py,v 1.11 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
This module defines gag functionality.
"""
import string
from lyntin import ansi, manager, utils, exported
from lyntin.modules import modutils

class GagData:
  def __init__(self):
    self._gags = {}
    self._antigags = {}

  def addGag(self, item):
    """
    Adds a gag to the dict.

    @param item: the item to gag
    @type  item: string
    """
    compiled = utils.compile_regexp(item, 1)
    self._gags[item] = compiled

  def addAntiGag(self, item):
    """ Adds an antigag."""
    compiled = utils.compile_regexp(item, 1)
    self._antigags[item] = compiled

  def clear(self):
    """
    Removes all the gags.
    """
    self._gags.clear()
    self._antigags.clear()

  def removeGags(self, text):
    """
    Removes gags from the list.

    Returns a list of tuples of gag item/gag that
    were removed.

    @param text: gags matching text will be removed
    @type  text: string

    @returns: list of (item, gag) tuples of removed gags
    @rtype: list of (string, string)
    """
    badgags = utils.expand_text(text, list(self._gags.keys()))

    ret = []
    for mem in badgags:
      ret.append(mem)
      del self._gags[mem]

    return ret

  def removeAntiGags(self, text=''):
    """
    Removes antigags from the list.

    @returns: a list of antigags that were removed.
    @rtype: list of strings
    """
    badgags = utils.expand_text(text, list(self._antigags.keys()))

    ret = []
    for mem in badgags:
      ret.append(mem)
      del self._antigags[mem]

    return ret

  def getAntiGags(self):
    """
    Returns the list of antigags that we have.

    @returns: list of all antigags
    @rtype: list of strings
    """
    listing = list(self._antigags.keys())
    listing.sort()
    return listing

  def expand(self, text):
    """
    Looks at mud data and performs any gags.

    It returns the final text--even if there were no gags.

    @param text: the text to expand gags in
    @type  text: string

    @return: the (un)adjusted text
    @rtype: string
    """
    if len(text) > 0:
      # check for antigags first
      for mem in list(self._antigags.values()):
        if mem.search(ansi.filter_ansi(text)):
          return text

      # check for gags
      for mem in list(self._gags.values()):
        if mem.search(ansi.filter_ansi(text)):
          tokens = ansi.split_ansi_from_text(text)
          tokens = [m for m in tokens if ansi.is_color_token(m)]
          return "".join(tokens)

    return text 

  def getInfo(self, text=''):
    """
    Returns information about the gags in here.

    This is used by #gag to tell all the gags involved
    as well as #write which takes this information and dumps
    it to the file.

    @param text: the text used to figure out which gags to provide
        information on
    @type  text: string

    @return: list of strings where each string represents a gag
    @rtype: list of strings
    """
    data = list(self._gags.keys())
    if text:
      data = utils.expand_text(text, data)

    data = ["gag {%s}" % mem for mem in data]

    return data

  def getGagInfoMappings(self):
    l = []
    for mem in list(self._gags.keys()):
      l.append( {"text": mem} )

    return l

  def getAntiGagsInfo(self, text=""):
    """
    Returns information about the antigags in here.

    This is used by #antigag to tell all the antigags involved
    as well as #write which takes this information and dumps it to the file.

    @param text: the text used to figure out which antigags to provide
        information on
    @type  text: string

    @return: list of strings where each string represents a gag
    @rtype: list of strings
    """
    data = list(self._antigags.keys())
    if text:
      data = utils.expand_text(text, data)

    data = ["antigag {%s}" % mem for mem in data]

    return data

  def getAntiGagInfoMappings(self):
    l = []
    for mem in list(self._antigags.keys()):
      l.append( { "item": mem } )

    return l

  def getStatus(self):
    """
    Returns a one liner of number of gags we're managing.

    @returns: string describing how many gags and gags we're managing
    @rtype: string
    """
    gags = len(list(self._gags.keys()))
    antigags = len(list(self._antigags.keys()))

    return "%d gag(s). %d antigag(s)" % (gags, antigags)


class GagManager(manager.Manager):
  def __init__(self):
    self._gagdata = {}

  def getGagData(self, ses):
    if ses not in self._gagdata:
      self._gagdata[ses] = GagData()
    return self._gagdata[ses]
    
  def clear(self, ses):
    if ses in self._gagdata:
      self._gagdata[ses].clear()

  def getInfo(self, ses, text=''):
    return self.getGagData(ses).getInfo(text)

  def getItems(self):
    return ["gag", "antigag"]

  def getParameters(self, item):
    if item == "gag":
      return [ ( "text", "The text whose presence indicates we should gag the line." ) ]

    if item == "antigag":
      return [ ( "item", "The item whose presence indicates we shouldn't gag the line." ) ]

    raise ValueError("%s is not a valid item for this manager." % item)

  def getInfoMappings(self, item, ses):
    if item not in ["gag", "antigag"]:
      raise ValueError("%s is not a valid item for this manager." % item)

    if ses not in self._gagdata:
      return []

    if item == "gag":
      return self._gagdata.getGagInfoMappings()

    return self._gagdata.getAntiGagInfoMappings()

  def getStatus(self, ses):
    return self.getGagData(ses).getStatus()

  def addSession(self, newsession, basesession=None):
    if basesession:
      if basesession in self._gagdata:
        bdata = self.getGagData(basesession)
        ndata = self.getGagData(newsession)

        for mem in list(bdata._gags.keys()):
          ndata.addGag(mem)
        for mem in list(bdata._antigags.keys()):
          ndata.addAntiGag(mem)

  def removeSession(self, ses):
    if ses in self._gagdata:
      del self._gagdata[ses]

  def persist(self, args):
    """
    write_hook function for persisting the state of our session.
    """
    ses = args["session"]
    quiet = args["quiet"]

    gd = self.getGagData(ses)
    data = gd.getInfo() + gd.getAntiGagsInfo()

    if quiet == 1:
      data = [m + " quiet={true}" for m in data]

    return data

  def mudfilter(self, args):
    """
    mud_filter_hook function to perform gagstitutions on data 
    that comes from the mud.
    """
    ses = args["session"]
    text = args["dataadj"]

    if exported.get_config("ignoresubs", ses, 0) == 0 and ses in self._gagdata:
      text = self._gagdata[ses].expand(text)
    return text


commands_dict = {}

def antigag_cmd(ses, args, input):
  """
  Allows you to create antigags.

  For any line that contains an antigag, we won't do gags on it.

  category: commands
  """
  item = args["item"]
  quiet = args["quiet"]

  gm = exported.get_manager("gag")
  gd = gm.getGagData(ses)

  if not item:
    data = gd.getAntiGagsInfo()
    if not data:
      data = ["antigag: no antigags defined."]

    exported.write_message("antigags\n" + "\n".join(data), ses)
    return

  gd.addAntiGag(item)
  if not quiet:
    exported.write_message("antigag: {%s} added." % item, ses)

commands_dict["antigag"] = (antigag_cmd, "item= quiet:boolean=false")

def unantigag_cmd(ses, args, input):
  """
  Allows you to remove antigags.

  category: commands
  """
  str = args["str"]
  quiet = args["quiet"]

  gm = exported.get_manager("gag")
  gd = gm.getGagData(ses)

  func = gd.removeAntiGags
  modutils.unsomething_helper(args, func, None, "antigag", "antigags")

commands_dict["unantigag"] = (unantigag_cmd, "str= quiet:boolean=false")

def gag_cmd(ses, args, input):
  """
  With no arguments, prints out all gags.
  With arguments, creates a gag.

  Incoming lines from the mud which contain gagged text will
  be removed and not shown on the ui.

  Gags get converted to regular expressions.  Feel free to use
  regular expression matching syntax as you see fit.

  As with all commands, braces get stripped off and each complete
  argument creates a gag.  

  examples:
    #gag {has missed you.}    <-- will prevent any incoming line
                                  with "has missed you" to be shown.
    #gag missed               <-- will gag lines with "missed" in them.
    #gag {r[sven.*?dealt]i}   <-- will gag anything that matches the
                                  regexp "sven.*?dealt" and ignore
                                  case.

  category: commands
  """
  gaggedtext = args["text"]
  quiet = args["quiet"]

  gm = exported.get_manager("gag")
  gd = gm.getGagData(ses)

  if not gaggedtext:
    data = gd.getInfo()
    if not data:
      data = ["gag: no gags defined."]

    exported.write_message("gags\n" + "\n".join(data), ses)
    return

  gd.addGag(gaggedtext)
  if not quiet:
    exported.write_message("gag: {%s} added." % gaggedtext, ses)

commands_dict["gag"] = (gag_cmd, "text= quiet:boolean=false")


def ungag_cmd(ses, args, input):
  """
  Allows you to remove gags.

  category: commands
  """
  str = args["str"]
  quiet = args["quiet"]

  gm = exported.get_manager("gag")
  gd = gm.getGagData(ses)

  func = gd.removeGags
  modutils.unsomething_helper(args, func, None, "gag", "gags")

commands_dict["ungag"] = (ungag_cmd, "str= quiet:boolean=false")


gm = None

def load():
  """ Initializes the module by binding all the commands."""
  global gm
  modutils.load_commands(commands_dict)
  gm = GagManager()
  exported.add_manager("gag", gm)

  exported.hook_register("mud_filter_hook", gm.mudfilter, 50)
  exported.hook_register("write_hook", gm.persist)

def unload():
  """ Unloads the module by calling any unload/unbind functions."""
  global gm
  modutils.unload_commands(list(commands_dict.keys()))
  exported.remove_manager("gag")

  exported.hook_unregister("mud_filter_hook", gm.mudfilter)
  exported.hook_unregister("write_hook", gm.persist)

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: constants.py,v 1.5 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
Holds constants which will get used through the code-base as well as in
various Lyntin modules.  We attempt to keep all the useful stuff in three
different modules: constants, exported, and utils.
"""
import lyntin

FIRST = 1
LAST = 99

TRUE_VALUES = ["yes", "true", "1", "on", 1]
FALSE_VALUES = ["no", "false", "0", "off", 0]

# lyntin's title and catch phrase
LYNTINTITLE = "Lyntin -- The Hacker's Mudclient "

# version information
VERSION = """Lyntin """ + lyntin.__version__ + """
For bugs, suggestions, mailing list info, feature requests,
architecture docs, et al, see """ + lyntin.__url__ + """
"""

# help text which gets printed to stdout if you do 'lyntin.py --help'
HELPTEXT = """syntax: lyntin.py [[OPTIONS] | [--help] | [--version]]

  --help
       displays this text and exits.

  -v or --version
       prints out the version information and exits.

  -c or --configuration
       sets the configuration file to use for setting up the
       datadir, moduledirs, plugins to load, files to read,
       ui to use, and other boot options.
"""

# Lyntin displays this after it's done initializing and it's
# ready for user interaction.
STARTUPTEXT = """Initialization complete.
-------------------------------------------------------------
Welcome to Lyntin.

Lyntin is licensed under the GPLv3.  For more details see
LICENSE.

%s
For help, type "#help general".
-------------------------------------------------------------
""" % VERSION

NODEFAULTVALUE = "No default value"



#########################################################################
# This file is part of Lyntin.
#
# Lyntin is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# Lyntin is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# copyright (c) Free Software Foundation 2001-2007
#
# $Id: commandmanager.py,v 1.8 2007/07/24 00:39:03 willhelm Exp $
#########################################################################
"""
Lyntin comes with a series of X{command}s for manipulating aliases, 
actions, highlights, and various other things.  Both the commands and 
the managers that hold the data are all defined in various Lyntin 
modules in the "modules/" subdirectory and are loaded up at Lyntin 
startup.

Lyntin has an extremely powerful argument parser that allows Lyntin
module writers to worry about the functionality of their command without
having to deal with parsing out the arguments, type-checking, 
raising exceptions with command syntax help, and data conversion.

Commands are stored internally in the "CommandManager".  They are
global to Lyntin--there are no commands that only exist in the session
they were declared in.  Commands, however, are executed in a specific
session.  When they are executed, they are given three arguments:
the session object for the session they were executed in, an argument
dictionary from the argument parser, and the exact input the user
typed.

Look at the modules/lyntincmds.py and modules/tintincmds.py modules for 
command examples.  Additionally, check out the Lyntin module repository
on http://lyntin.sourceforge.net/ for more examples.

X{default_resolver_hook}::

   Allows you to fill in values for arguments the user didn't specify 
   on the command line.  

   Arg mapping: {"session": Session, "commandname", string}

   session - the session this command was executed in.

   commandname - the name of the command that was executed

"""
import inspect, re
from lyntin import manager, exported, argparser, utils

class _CommandData:
  """
  Holds data relating to a command.  It's a helper class.
  """
  def __init__(self):
    self._name = ""
    self._func = None
    self._argparser = None
    self._fqn = ""

  def __repr__(self): return self._name
  def __str__(self): return self._name

  def setName(self, name): self._name = name
  def getName(self): return self._name
  def setNameAdjusted(self, name): self._name_adjusted = name
  def getNameAdjusted(self): return self._name_adjusted
  def setFunc(self, func): self._func = func
  def getFunc(self): return self._func
  def setArgParser(self, ap): self._argparser = ap
  def getArgParser(self): return self._argparser
  def setFQN(self, fqn): self._fqn = fqn
  def getFQN(self): return self._fqn

class CommandManager(manager.Manager):
  """ 
  The CommandManager holds a series of _CommandData objects
  and methods to manipulate and use them.  Lyntin developers
  can add their own commands to Lyntin.

  There should be one instance of the CommandManager and
  the engine should have it.  All CommandManager interaction
  should be done through the exported module.
  """
  def __init__(self, e):
    self._commands = {}
    self._engine = e

  def getCommands(self):
    """
    Returns a list of the commands we have registered.

    @return: all the commands that have been registered
    @rtype:  list of strings
    """
    return list(self._commands.keys())

  def addCommand(self, name, func, arguments=None, argoptions=None, helptext=""):
    """
    Registers a command.

    @param name: the name of the command to add
    @type  name: string

    @param func: the function that handles the command
    @type  func: function

    @param arguments: the argument spec to create the argparser
    @type  arguments: string

    @param argoptions: options for how the argument spec should be parsed
    @type  argoptions: string

    @param helptext: the help text for this command for the user
    @type  helptext: string

    @raise ValueError: if the function is uncallable
    @raise Exception: if the argument spec for the command is unparseable
    """
    if not callable(func):
      raise ValueError("%s is uncallable." % name)
    cd = _CommandData()

    syntaxline = ""

    # try to figure out the arguments and syntax line stuff
    if arguments is not None:
      try:
        cd.setName(name)
        cd.setArgParser(argparser.ArgumentParser(arguments, argoptions))
        syntaxline = utils.wrap_text(cd.getArgParser().syntaxline, 60, 6)
      except Exception as e:
        raise Exception("Error with arguments for command %s, (%s)" % (name,e))

    cd.setFunc(func)

    # removeCommand tests to see if the command exists already and will
    # remove it if it does.
    self.removeCommand(name)

    # toss the command thing in the dict
    self._commands[name] = cd

    # deal with the help text
    if not helptext:
      if func.__doc__:
        helptext = inspect.getdoc(func)
      else:
        helptext = "\nThis command has no help."

    if name.startswith("^"):
      cd.setNameAdjusted(name[1:])
    else:
      cd.setNameAdjusted(name)

    if syntaxline:
      commandchar = self._engine.getConfigManager().get("commandchar")
      helptext = ("syntax: %s%s %s\n" % 
             (commandchar, cd.getNameAdjusted(), syntaxline) + helptext)

    fqn = exported.add_help(cd.getNameAdjusted(), helptext)
    cd.setFQN(fqn)
        
  def removeCommand(self, name):
    """
    Removes a command (and the help text) for whatever reasons.

    @param name: the name of the command to remove
    @type  name: string

    @return: 0 if no command was found, 1 if the command was removed
        succesfully.
    @rtype: boolean
    """
    if name in self._commands:
      cd = self._commands[name]
      del self._commands[name]
      try:
        exported.remove_help(cd.getFQN())
      except:
        pass
      return 1
    return 0

  def getCommand(self, name):
    """
    Returns the function for a given command name.

    @param name: the name of the command to retrieve
    @type  name: string

    @return: the function in question or None
    @rtype:  function
    """
    if name in self._commands:
      return self._commands[name].getFunc()

    if ("^" + name) in self._commands:
      return self._commands["^" + name].getFunc()

    # this is kind of a kluge to handle the #@ arbitrary
    # python stuff so that it can be in its own module.
    if name.startswith("@") and "@" in self._commands:
      return self._commands["@"].getFunc()

    return None

  def getArgParser(self, name):
    """
    Returns the arguments parser for a given command name.

    @param name: the name of the command whose arguments should be
        retrieved
    @type  name: string

    @return: argument parsing object to convert incoming arguments
        into a dictionary to pass to the command function
    @rtype: ArgParser instance
    """
    if name in self._commands:
      return self._commands[name].getArgParser()

    return None

  def filter(self, args):
    """
    Takes in user command lines and handles commands that start
    with a Lyntin command character.

    @param args: (session, internal boolean, ..., current input text)
    @type  args: tuple

    @return: None if we handled the input, or the current input text if 
        we didn't
    @rtype:  None or string
    """
    ses = args["session"]
    internal = args["internal"]
    input = args["dataadj"]

    commandchar = self._engine.getConfigManager().get("commandchar")
    if len(input) > 1 and input.startswith(commandchar):
      input = input[1:]

      # splits out the command name from the rest of the command line
      words = input.split(" ",1)

      # We want an empty argument list if there was one, don't want
      # array out-of-bounds issues       
      if len(words) < 2: words.append("")

      # this checks to see if it's a special #@ command.
      if input.startswith("@"):
        self.getCommand("@")(ses, input.split(" "), input)
        if internal==0: ses.prompt()
        return

      # this finds the first matching command and ends there.
      commands = self.getCommands()
      for mem in commands:
        command = None
        if mem.startswith("^"):
          if re.compile(mem).search(words[0]):
            command = self.getCommand(mem)
        else:
          if mem.startswith(words[0]):
            command = self.getCommand(mem)

        if command:
          argumentparser = self.getArgParser(mem)
          if argumentparser is None:
            command(ses, input.split(" "), input)
          else:
            # for printing out the error message, we remove the ^
            # from the command name if it's there.
            fixedmem = mem
            if len(fixedmem) > 0 and fixedmem.startswith("^"):
              fixedmem = fixedmem[1:]

            resolver = exported.hook_spam("default_resolver_hook", 
                                {"session": ses, "commandname": mem}, 
                                mappingfunc=exported.query_mapper, 
                                donefunc=exported.query_done)

            try:
              argdict = argumentparser.parse(words[1], resolver)
              argdict["command"]=mem
              command(ses, argdict, input)
            except ValueError as e:
              exported.write_error("%s: %s\nsyntax: %s%s %s" % 
                                   (fixedmem, e, commandchar, fixedmem,
                                    argumentparser.syntaxline))
            except argparser.ParserException as e:
              exported.write_error("%s: %s\nsyntax: %s%s %s" % 
                                   (fixedmem, e, commandchar, fixedmem,
                                    argumentparser.syntaxline))
          if internal == 0:
            ses.prompt()
          break

      else:
        if internal == 0:
          ses.prompt()
        exported.write_error("Not a valid command: %s" % (words[0]))
      return
    return args["dataadj"]

# Local variables:
# mode:python
# py-indent-offset:2
# tab-width:2
# End:
